# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is used to define the process of the reference creator.
This is related to the issue #184
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "05/02/2018"


import os
import shutil
import tempfile
import h5py
import fabio
import numpy
import typing
from silx.io.url import DataUrl
from tomwer.core import utils
from silx.io.dictdump import dicttoh5, h5todict
from tomoscan.io import HDF5File
import logging
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefsWorker, DarkRefs
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.signal import Signal
from tomwer.core.process.baseprocess import _input_desc, _output_desc
from .settings import DARKHST_PREFIX, REFHST_PREFIX

logger = logging.getLogger(__name__)


class DarkRefsCopy(DarkRefs):
    """
    Reimplement Dark ref to deal with copy when there is no median/mean files
    """

    inputs = [
        _input_desc(
            name="data", type=TomwerScanBase, handler="process", doc="scan path"
        )
    ]

    outputs = [_output_desc(name="data", type=TomwerScanBase, doc="scan path")]

    def __init__(self, reconsparams=None):
        super(DarkRefsCopy, self).__init__(reconsparams=reconsparams)
        self.refPrefix = REFHST_PREFIX
        self.darkPrefix = DARKHST_PREFIX

        # API exposition
        self.set_mode_auto = self.worker.set_mode_auto
        self.has_flat_stored = self.worker.has_flat_stored
        self.has_dark_stored = self.worker.has_dark_stored
        self.has_flat_or_dark_stored = self.worker.has_flat_or_dark_stored
        self.set_refs_from_scan = self.worker.set_refs_from_scan
        self.get_refHST_prefix = self.worker.refHST_prefix
        self.is_on_mode_auto = self.worker.is_on_mode_auto
        self.set_process_only_dkRf = self.worker.set_process_only_dkRf
        self.set_process_only_copy = self.worker.set_process_only_copy
        self.clear_ref = self.worker.clear_ref

    def _createThread(self):
        return DarkRefsCopyWorker(process=self)

    def setModeAuto(self, auto):
        self.worker.set_mode_auto(auto)

    def set_properties(self, properties):
        super().set_properties(properties=properties)

    def get_dark_save_file(self):
        return self.worker._dark_save_file

    def get_ref_save_file(self):
        return self.worker._ref_save_file


class DarkRefsCopyWorker(DarkRefsWorker):
    """
    Add to DarkRef the copy of reference if not existing in the given directory
    """

    DEFAULT_SRCURRENT = 200.0  # mA

    sigRefSetted = Signal(TomwerScanBase)
    """Signal emitted when some reference are recorded from the given directory
    """
    sigRefRemoved = Signal()
    """Signal emitted when the reference are removed"""

    def __init__(self, process, *args, **kwargs):
        DarkRefsWorker.__init__(self, process=process, *args, **kwargs)
        self._modeAuto = True
        self._savedir = tempfile.mkdtemp()
        self._dark_save_file = os.path.join(self._savedir, "dark.h5")
        self._ref_save_file = os.path.join(self._savedir, "refs.h5")
        info = "Ref copy will store files in " + self._savedir
        logger.info(info)
        """directory saved to store the dark and refHST reference files"""
        self._refHstPrefix = REFHST_PREFIX
        self._darkHstPrefix = DARKHST_PREFIX
        self._processOnlyCopy = False
        self._processOnlyDkRf = False

    def set_process_only_dkRf(self, value: bool) -> None:
        self._processOnlyDkRf = value
        self._processOnlyCopy = False

    def set_process_only_copy(self, value: bool) -> None:
        self._processOnlyDkRf = False
        self._processOnlyCopy = value

    def set_refs_from_scan(self, scan: TomwerScanBase) -> bool:
        assert isinstance(scan, TomwerScanBase)
        if isinstance(scan, HDF5TomoScan):
            darks = DarkRefs.get_darks_frm_process_file(
                scan.process_file, entry=scan.entry, as_url=True
            )
            if len(darks) == 0:
                logger.warning("No dark found in %s")
            flats = DarkRefs.get_flats_frm_process_file(
                scan.process_file, entry=scan.entry, as_url=True
            )
            if len(flats) == 0:
                logger.warning("No flat found in %s")
            has_flat_or_dark = len(darks) > 0 or len(flats) > 0
            if not has_flat_or_dark:
                self._notice_user_no_ref_message(fileType="flat field", dir=scan)
                return False
            else:
                self.clean_save_files()
                if len(darks) == 0:
                    self._notice_user_no_ref_message(fileType="dark", dir=scan)
                else:
                    key_0 = sorted(darks.keys())[0]
                    self._copy_dark_from(dark_url=darks[key_0], scan=scan)

                if len(flats) == 0:
                    self._notice_user_no_ref_message(fileType="flat field", dir=scan)
                else:
                    key_0 = sorted(flats.keys())[0]
                    self._copy_ref_from(url=flats[key_0], scan=scan)
                self.sigRefSetted.emit(scan)
                return True
        elif isinstance(scan, EDFTomoScan):
            if os.path.isdir(scan.path) is False:
                w = "given path (%s) is not a directory. Can't extract ref" % scan
                logger.warning(w)
                return False

            self.directory = scan
            if not self.contains_dark_or_ref(scan):
                self._notice_user_no_ref_message(fileType="flat field", dir=scan)
                return False
            else:
                self.clean_save_files()
                darkHSTFiles = sorted(
                    DarkRefs.getDarkHSTFiles(scan.path, prefix=self.darkHST_prefix)
                )
                refHSTFiles = sorted(
                    DarkRefs.getRefHSTFiles(scan.path, prefix=self.refHST_prefix)
                )
                if len(darkHSTFiles) == 0:
                    self._notice_user_no_ref_message(fileType="dark", dir=scan)
                else:
                    self._copy_dark_from(
                        dark_url=DataUrl(file_path=darkHSTFiles[0], scheme="fabio"),
                        scan=scan,
                    )

                if len(refHSTFiles) == 0:
                    self._notice_user_no_ref_message(fileType="flat field", dir=scan)
                else:
                    self._copy_ref_from(
                        url=DataUrl(file_path=refHSTFiles[0], scheme="fabio"), scan=scan
                    )

                """warning: dark should always be saved first because used for
                saving ref
                """
                self.sigRefSetted.emit(scan)
                return True

    def _copy_to(self, scan):
        assert isinstance(scan, TomwerScanBase)

        if isinstance(scan, EDFTomoScan):
            if os.path.isdir(scan.path) is False:
                logger.warning(
                    str(scan) + " is not a directory. Cannot copy " "reference to it."
                )
                return
            if self.has_flat_or_dark_stored() is False:
                logger.error("No reference registred to be copy in %s" % scan)
                return

        # warning: get information if there is ref or dark before copying
        # because can bring interferences.
        _hasRef = self.contains_flat(scan)
        _hasDark = self.contains_dark(scan)
        # special treatment for dark and ref we want to store both at the same
        # time
        if isinstance(scan, HDF5TomoScan) and not _hasDark and not _hasRef:
            self._copy_dark_and_flat_to(scan)
        else:
            if self.has_dark_stored() and not _hasDark:
                self._copy_dark_to(scan)
            if self.has_flat_stored() and not _hasRef:
                self._copy_flat_to(scan)

    def _copy_dark_and_flat_to(
        self, scan: HDF5TomoScan, copy_dark: bool = True, copy_flats: bool = True
    ):
        """
        Function dedicated to the copy of dark and/or flat to HDF5TomoScan
        :param HDF5TomoScan scan:
        :param bool copy_dark:
        :param bool copy_flats:
        :return:
        """
        assert isinstance(scan, HDF5TomoScan), "function dedicated to HDF5TomoScan"
        if copy_dark is False and copy_flats is False:
            return

        results = {}
        shape_dark = None
        shape_flat = None
        if copy_dark:
            dark_info = self._get_dark_info()
            if dark_info is not None:
                results["darks"] = {"0": dark_info["data"]}
                shape_dark = dark_info["data"].shape
                scan.set_normed_darks(results["darks"])
        if copy_flats:
            flat_info = self._get_flat_info()
            if flat_info is not None:
                flats = {"0": flat_info["data_start"]}
                shape_flat = flat_info["data_start"].shape
                index_last_projection = max(scan.projections) or None
                if index_last_projection is not None and index_last_projection != 0:
                    flats[str(index_last_projection)] = flat_info["data_end"]
                results["flats"] = flats
                scan.set_normed_flats(results["flats"])

        # check dimensions of dark and flat compare to projections if any
        if scan.dim_1 is not None and scan.dim_2 is not None:
            shape_frame = (scan.dim_2, scan.dim_1)
            incoherent_with_dark = False
            incoherent_with_flat = False
            if shape_dark is not None:
                incoherent_with_dark = shape_dark != shape_frame
            if shape_flat is not None:
                incoherent_with_flat = shape_flat != shape_frame

            if incoherent_with_dark:
                logger.warning(
                    "Scan frame (%s) and dark (%s) to copy have"
                    "different shapes. Unable to do the copy"
                    "" % (shape_frame, shape_dark)
                )
            if incoherent_with_flat:
                logger.warning(
                    "Scan frame (%s) and flat (%s) to copy have"
                    "different shapes. Unable to do the copy"
                    "" % (shape_frame, shape_flat)
                )
            if incoherent_with_flat or incoherent_with_dark:
                return

        with scan.acquire_process_file_lock():
            DarkRefs._register_process(
                process_file=scan.process_file,
                entry=scan.entry,
                process=DarkRefs,
                configuration=self.recons_params.to_dict(),
                results=results,
                process_index=self.scan.pop_process_index(),
                overwrite=True,
            )

    def _copy_dark_to(self, scan):
        """The dark is copied without normalization"""
        assert isinstance(scan, TomwerScanBase)
        if not self.has_dark_stored():
            raise ValueError("require to copy dark but no dark stored")
        dark_info = self._get_dark_info()
        assert dark_info is not None, (
            "when this function is call you should insure"
            "that process has stored some dark"
        )
        if isinstance(scan, EDFTomoScan):
            # do never copy on an existing file (meaning ref already here)
            dark_file_name = dark_info["dark_basename"] or DARKHST_PREFIX
            dst = os.path.join(scan.path, os.path.basename(dark_file_name))
            if os.path.isfile(dst) is True:
                logger.info("%s exists, will not copy dark" % dst)
            else:
                header = dark_info["header"]
                data = dark_info["data"]
                file_desc = fabio.edfimage.EdfImage(data=data, header=header)
                file_desc.write(dst, force_type=numpy.int32)

        elif isinstance(scan, HDF5TomoScan):
            self._copy_dark_and_flat_to(scan=scan, copy_dark=True, copy_flats=False)
        else:
            raise ValueError("Tomo scan type not managed (%s)" % scan)

    def _copy_flat_to(self, scan):
        assert isinstance(scan, TomwerScanBase)
        if not self.has_flat_stored():
            raise ValueError("require to copy flat but no flat stored")
        if isinstance(scan, EDFTomoScan):

            def normalize(data, when):
                """normalize from dark_end and getting the SRCurrent at start or
                end"""
                assert when in ("start", "end")

                srCurrent = utils.getSRCurrent(scan_dir=scan.path, when=when)
                if srCurrent is None:
                    logger.warning(
                        "Can't find information about srCurrent," "set to default value"
                    )
                    srCurrent = self.DEFAULT_SRCURRENT
                dark_info = self._get_dark_info()
                if dark_info is None:
                    logger.info("no dark found, unable to normalize")
                    return None, srCurrent
                dark_data = dark_info["data"]
                if data.shape != dark_data.shape:
                    logger.warning(
                        "Image and dark have different shapes."
                        "Cannot normalize. (Dark ref file in %s)"
                        "" % self._dark_save_file
                    )
                    return data, srCurrent
                return (data * srCurrent).astype(numpy.float32) + dark_data, srCurrent

            flat_info = self._get_flat_info()
            if flat_info is None:
                raise RuntimeError("No ref found during processing")
            tomo_N = utils.getTomo_N(scan.path)
            if tomo_N in (None, -1):
                logger.error(
                    "Can't find the number of projection. " "Fail to create reference"
                )

            header = flat_info["header"]
            end_acqui = str(utils.getTomo_N(scan.path)).zfill(4)
            indexRefFile = {"start": "0000", "end": end_acqui}
            for when in ("start", "end"):
                data = flat_info["_".join(("data", when))]
                fileName = self.recons_params.ref_prefix + indexRefFile[when] + ".edf"
                filePath = os.path.join(scan.path, fileName)
                # do never copy on an existing file (meaning ref already here)
                if os.path.isfile(filePath):
                    continue
                _data, _srCurrent = normalize(data, when)

                _header = header.copy()
                # add some extra information on the header
                _header["SRCUR"] = _srCurrent
                file_desc = fabio.edfimage.EdfImage(data=_data, header=_header)
                file_desc.write(filePath, force_type=numpy.int32)
        elif isinstance(scan, HDF5TomoScan):
            self._copy_dark_and_flat_to(scan=scan, copy_dark=False, copy_flats=True)
        else:
            raise ValueError("Tomo scan type not managed (%s)" % scan)

    def _notice_user_no_ref_message(self, fileType, dir):
        logger.warning(self._get_no_ref_message(fileType, dir))

    def _get_no_ref_message(self, fileType, dir):
        mess = (
            "No %s found in the given directory %s. Won't be able to copy"
            " them" % (fileType, dir)
        )
        return mess

    def _copy_ref_from(self, scan, url):
        """
        Copy the data contained in _file in the given file to the `_savedir`
        and normalize the data from SRCurrent (intensity
        )"""

        def write(data_start, data_end, header):
            with HDF5File(self._ref_save_file, "w", swmr=True) as h5f:
                h5f["data_start"] = data_start
                h5f["data_end"] = data_end
                h5f["tomwer_info"] = "copied using refCopy"
                h5f["original_scan"] = str(self.scan)
                h5f["srcurrent"] = -1
            # update header if any edf write
            header["tomwer_info"] = "copied using refCopy"
            header["original_scan"] = str(self.scan)
            header["SRCUR"] = -1
            dicttoh5(
                header,
                h5file=self._ref_save_file,
                h5path="header",
                overwrite_data=False,
                mode="a",
            )

        assert isinstance(url, DataUrl)
        if url.scheme() == "fabio":

            def normalize_edf(data):
                srCurrent = utils.getClosestSRCurrent(
                    scan_dir=scan.path, refFile=url.file_path()
                )
                if srCurrent in (None, -1):
                    logger.warning(
                        "Can't find information about srCurrent,"
                        "set to default value for normalization"
                    )
                    srCurrent = self.DEFAULT_SRCURRENT

                if not self.contains_dark(scan=scan):
                    logger.warning("No darkHST recorded, unable to normalize")
                    return data, srCurrent

                dark_data = self._get_dark_info()
                if dark_data is None:
                    logger.warning("No darkHST found in %s" % self._dark_save_file)
                    return None, None
                else:
                    dark_data = dark_data["data"]
                if data.shape != dark_data.shape:
                    err = (
                        "cannot normalize data from %s, has different "
                        "dimensions" % (self._darkHST)
                    )
                    logger.error(err)
                    return data, srCurrent
                return (data - dark_data).astype(numpy.float32) / srCurrent, srCurrent

            with fabio.open(url.file_path()) as file_desc:
                data = file_desc.data
                header = file_desc.header
                data, sr_current = normalize_edf(data)
                write(data_start=data, data_end=data, header=header)
                # TODO: look but the writing should be more or less generic
        elif url.scheme() in ("silx", "h5py"):
            flats = DarkRefs.get_flats_frm_process_file(
                scan.process_file, entry=scan.entry, as_url=False
            )
            if len(flats) == 0:
                logger.warning(
                    "No flat found in %s, unable to copy" % scan.process_file
                )
                return
            key_0 = sorted(list(flats.keys()))[0]
            data_start = flats[key_0]
            if len(flats) > 2:
                key_end = sorted(list(flats.keys()))[-1]
                data_end = flats[key_end]
            else:
                data_end = data_start
            dark_file_basename = None
            # TODO: get some metadata ?
            header = {}
            write(data_start=data_start, data_end=data_end, header=header)
        else:
            raise ValueError("scheme not managed")

    def _copy_dark_from(self, scan, dark_url: DataUrl):
        assert isinstance(dark_url, DataUrl)
        assert isinstance(scan, TomwerScanBase)
        if dark_url.scheme() == "fabio":
            with fabio.open(dark_url.file_path()) as dsc:
                header = dsc.header.copy()
                data = dsc.data.copy()
                dark_file_basename = os.path.basename(dark_url.file_path())
        elif dark_url.scheme() in ("h5py", "silx"):
            darks = DarkRefs.get_darks_frm_process_file(
                scan.process_file, entry=scan.entry, as_url=False
            )
            if len(darks) == 0:
                logger.warning(
                    "No dark found in %s, unable to copy" % scan.process_file
                )
                return
            key_0 = sorted(list(darks.keys()))[0]
            data = darks[key_0]
            dark_file_basename = None
            # TODO: get some metadata ?
            header = {}
        else:
            raise ValueError("scheme not managed")

        # add origin of the dark file
        with HDF5File(self._dark_save_file, "w", swmr=True) as h5f:
            # h5f['header'] = header
            h5f["tomwer_info"] = "copied using refCopy"
            h5f["original_scan"] = str(scan)
            h5f["data"] = data
            h5f["srcurrent"] = -1
            if dark_file_basename is not None:
                h5f["dark_basename"] = dark_file_basename
        header["tomwer_info"] = "copied using refCopy"
        header["original_scan"] = str(self.scan)
        header["SRCUR"] = -1
        dicttoh5(
            header,
            h5file=self._dark_save_file,
            h5path="header",
            overwrite_data=False,
            mode="a",
        )

    def _get_flat_info(self) -> typing.Union[None, dict]:
        """return ref data. Only one store for now"""
        if self.has_flat_stored():
            with HDF5File(self._ref_save_file, "r", swmr=True) as h5f:
                return {
                    "data_start": h5f["data_start"][()],
                    "data_end": h5f["data_end"][()],
                    "original_scan": h5f["original_scan"][()],
                    "srcurrent": h5f["srcurrent"][()],
                    "header": h5todict(h5file=self._dark_save_file, path="header"),
                    "tomwer_info": h5f["tomwer_info"][()],
                }
        else:
            return None

    def _get_dark_info(self) -> typing.Union[None, dict]:
        """return dark data, header, original_scan"""
        if self.has_dark_stored():
            with HDF5File(self._dark_save_file, "r", swmr=True) as h5f:
                if "dark_basename" in h5f:
                    dark_basename = h5f["dark_basename"][()]
                else:
                    dark_basename = None
                res = {
                    "data": h5f["data"][()],
                    "original_scan": h5f["original_scan"][()],
                    "srcurrent": h5f["srcurrent"][()],
                    "dark_basename": dark_basename,
                    "tomwer_info": h5f["tomwer_info"][()],
                    "header": h5todict(h5file=self._dark_save_file, path="header"),
                }
                return res
        else:
            return None

    def __del__(self):
        if os.path.exists(self._savedir):
            shutil.rmtree(self._savedir)

    def clean_save_files(self):
        for f in os.listdir(self._savedir):
            if os.path.isfile(f) and os.path.exists(f):
                os.remove(f)

    # TODO: should take scan as input ? !!!
    def process(self):
        """
        This is function triggered when a new scan / data is received.
        As explained in issue #184 the behavior is the following:

        * if the scan has already ref files files won't be overwrite
        * if the mode is in `auto` will register last ref file met
        * if the scan has no ref files and refCopy has some register. Will
          create a copy of those, normalized from srCurrent (for flat field)
        """
        if self.scan is None or self.scan.path is None:
            return
        if not self._processOnlyCopy:
            DarkRefsWorker.process(self)
        if not self._processOnlyDkRf:
            if self.contains_dark_or_ref(self.scan):
                if self._modeAuto:
                    self.set_refs_from_scan(self.scan)
            if self.has_missing_dark_or_ref(self.scan):
                self._copy_to(self.scan)

    def has_flat_or_dark_stored(self) -> bool:
        """

        :return: True if the process has at least registered one flat or one
                 dark
        :rtype: bool
        """
        return self.has_flat_stored() or self.has_dark_stored()

    def has_flat_stored(self) -> bool:
        """

        :return: True if the process has registered at least one ref
        :rtype: bool
        """
        return os.path.exists(self._ref_save_file)

    def has_dark_stored(self) -> bool:
        """

        :return: True if the process has registered at least one dark
        :rtype: bool
        """
        return os.path.exists(self._dark_save_file)

    def contains_dark(self, scan: TomwerScanBase) -> bool:
        """Return True if the scan has already some dark processed"""
        assert isinstance(scan, TomwerScanBase)
        if isinstance(scan, EDFTomoScan):
            return (
                len(
                    DarkRefs.getDarkHSTFiles(
                        directory=scan.path, prefix=self.darkHST_prefix
                    )
                )
                > 0
            )
        elif isinstance(scan, HDF5TomoScan):
            if scan.process_file is None or not os.path.exists(scan.process_file):
                return False
            else:
                return (
                    len(
                        DarkRefs.get_darks_frm_process_file(
                            scan.process_file, entry=scan.entry, as_url=True
                        )
                    )
                    > 0
                )
        else:
            raise ValueError("scan type not managed")

    def contains_flat(self, scan: TomwerScanBase):
        """Return True if the scan has already some dark processed"""
        assert isinstance(scan, TomwerScanBase)
        if isinstance(scan, EDFTomoScan):
            return (
                len(
                    DarkRefs.getRefHSTFiles(
                        directory=scan.path, prefix=self.refHST_prefix
                    )
                )
                > 0
            )
        elif isinstance(scan, HDF5TomoScan):
            if scan.process_file is None or not os.path.exists(scan.process_file):
                return False
            else:
                return (
                    len(
                        DarkRefs.get_flats_frm_process_file(
                            scan.process_file, entry=scan.entry, as_url=True
                        )
                    )
                    > 0
                )
        else:
            raise ValueError("scan type not managed")

    def contains_dark_or_ref(self, scan):
        return self.contains_dark(scan=scan) or self.contains_flat(scan=scan)

    def has_missing_dark_or_ref(self, scan: TomwerScanBase) -> bool:
        """return True if the scan has no ref or no dark registered"""
        assert isinstance(scan, TomwerScanBase)
        return not self.contains_dark(scan) or not self.contains_flat(scan)

    def _signal_done(self, scan):
        assert isinstance(scan, TomwerScanBase)
        raise NotImplementedError("Abstract class")

    def set_mode_auto(self, b):
        self._modeAuto = b

    @property
    def is_on_mode_auto(self):
        return self._modeAuto

    @property
    def refHST_prefix(self):
        return self._refHstPrefix

    @property
    def darkHST_prefix(self):
        return self._darkHstPrefix

    def set_refHST_prefix(self, prefix):
        self._refHstPrefix = prefix

    def set_darkHST_prefix(self, prefix):
        self._darkHstPrefix = prefix

    def clear_ref(self):
        os.remove(self._ref_save_file)
        os.remove(self._dark_save_file)
        self.sigRefRemoved.emit()



# coding: utf-8
###########################################################################
# Copyright (C) 2016-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "03/05/2019"


from tomwer.core.process.reconstruction.axis import AxisProcess
from tomwer.core.process.reconstruction.axis.axis import NoAxisUrl
from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.synctools.axis import QAxisRP
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.utils import isLowOnMemory
from ..processingstack import FIFO, ProcessingThread
from silx.gui import qt
import logging
import functools

_logger = logging.getLogger(__name__)


class AxisProcessStack(FIFO, qt.QObject):
    """Implementation of the `.AxisProcess` but having a stack for treating
    scans and making computation in threads"""

    def __init__(self, axis_params):
        qt.QObject.__init__(self)
        FIFO.__init__(self)
        assert axis_params is not None
        self._axis_params = axis_params

    def _process(self, scan, configuration, callback=None):
        assert isinstance(scan, TomwerScanBase)
        if scan.axis_params is None:
            scan.axis_params = QAxisRP()
        self._scan_currently_computed = scan
        self._axis_params.frame_width = scan.dim_1
        mode = self._axis_params.mode
        if isOnLbsram(scan) and isLowOnMemory(get_lbsram_path()) is True:
            # if computer is running into low memory on lbsram skip it
            mess = "low memory, skip axis calculation", scan.path
            _logger.processSkipped(mess)
            scan.axis_params.set_value_ref_tomwer(None)
            if callback is not None:
                callback()
            self.scan_ready(scan=scan)
        elif not self._axis_params.use_sinogram and mode in (
            AxisMode.manual,
            AxisMode.read,
        ):
            scan._axis_params.set_value_ref_tomwer(self._axis_params.value_ref_tomwer)
            cor = scan._axis_params.value_ref_tomwer
            if isinstance(scan, HDF5TomoScan):
                entry = scan.entry
            else:
                entry = "entry"
            with scan.acquire_process_file_lock():
                AxisProcess._register_process(
                    process_file=scan.process_file,
                    entry=entry,
                    process=AxisProcess,
                    results={"center_of_rotation": cor if cor is not None else "-"},
                    configuration=self._axis_params.to_dict(),
                    process_index=scan.pop_process_index(),
                    overwrite=True,
                )
            # If mode is read or manual the position_value is not computed and
            # we will keep the actual one (should have been defined previously)
            self._end_computation(scan=scan, callback=callback)

        elif (
            not self._axis_params.use_sinogram
            and not mode in AxisProcess.RADIO_CALCULATIONS_METHODS
        ):
            _logger.warning("no method defined to compute {}".format(mode))
            if callback is not None:
                callback()
            self._process_next()
        else:
            scan.axis_params.set_value_ref_tomwer("...")
            self._axis_params.set_value_ref_tomwer("...")
            assert self._axis_params.value_ref_tomwer == "..."
            self._scan_currently_computed = scan
            self._computationThread.init(scan=scan, axis_params=self._axis_params)
            # need to manage connect before starting it because
            fct_callback = functools.partial(self._end_threaded_computation, callback)
            self._computationThread.finished.connect(fct_callback)
            self._computationThread.start()

    def _end_computation(self, scan, callback):
        """
        callback when the computation thread is finished

        :param scan: pass if no call to '_computationThread is made'
        """
        assert isinstance(scan, TomwerScanBase)
        assert self._axis_params is not None
        # copy result computed on scan on the AxisProcess reconsparams
        self._axis_params.set_value_ref_tomwer(
            scan.axis_params.value_ref_tomwer
        )  # noqa
        self._axis_params.frame_width = scan.dim_1
        FIFO._end_computation(self, scan=scan, callback=callback)

    def _end_threaded_computation(self, callback=None):
        assert self._scan_currently_computed is not None
        self._axis_params.set_value_ref_tomwer(
            self._computationThread.center_of_rotation
        )
        self._computationThread.finished.disconnect()
        if callback:
            callback()
        FIFO._end_threaded_computation(self)

    def _create_processing_thread(self) -> qt.QThread:
        return _ProcessingThread()


class _ProcessingThread(ProcessingThread):
    """
    Thread use to execute the processing of the axis position
    """

    def __init__(self):
        ProcessingThread.__init__(self)
        self._scan = None
        self._axis_params = None
        """function pointer to know which function to call for the axis
        calculation"""
        self.__patch = {}
        """Used to patch some calculation method (for test purpose)"""

    def init(self, scan, axis_params):
        self._scan = scan
        self._axis_params = axis_params

    def run(self):
        self.sigComputationStarted.emit()
        axis = AxisProcess(axis_params=self._axis_params)
        axis = self.apply_patch(axis=axis)
        try:
            axis.process(self._scan)
        except NoAxisUrl as e:
            self.center_of_rotation = None
            _logger.error(str(e))
        except Exception as e:
            _logger.error(str(e))
            self.center_of_rotation = None
        else:
            self.center_of_rotation = self._scan.axis_params.value_ref_tomwer

    def patch_calc_method(self, mode, function):
        self.__patch[mode] = function

    def apply_patch(self, axis):
        for mode, patch_fct in self.__patch.items():
            if mode in AxisMode:
                axis.RADIO_CALCULATIONS_METHODS[mode] = patch_fct
        return axis



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "31/07/2020"


import unittest
from tomwer.test.utils import UtilsTest
import h5py
import shutil
import tempfile
from tomwer.core.utils.scanutils import MockBlissAcquisition
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from pypushflow.representation.scheme.ows_parser import OwsParser
from tomwer.core.process.reconstruction.nabu import settings as nabu_settings
from ..process import _exec
from glob import glob
import os


class TestNxTomoDarkRef(unittest.TestCase):
    """
    Test the execution of the following workflow:
    nxtomomill -> dark and flat field
    uses:

    - 'nxtomomill_darkref.ows': where dark is not computed and ref is average
    - 'nxtomomill_darkref_2.ows': where dark is average and ref is median
    """

    TIMEOUT = 120

    def setUp(self) -> None:
        unittest.TestCase.setUp(self)
        self._data_dir = tempfile.mkdtemp()
        self._bliss_mock = MockBlissAcquisition(
            n_sample=1,
            n_sequence=1,
            n_darks=10,
            n_flats=10,
            n_scan_per_sequence=2,
            output_dir=self._data_dir,
        )
        self.master_file = self._bliss_mock.samples[0].sample_file
        assert os.path.exists(self.master_file)
        self.bliss_scan = BlissScan(
            master_file=self.master_file, entry="/1.1", proposal_file=None
        )
        self.tomwer_file = os.path.join(
            self._bliss_mock.samples[0].path, "tomwer_processes.h5"
        )

    @staticmethod
    def get_scheme(file_path):
        return OwsParser.scheme_load(file_path, load_handlers=True)

    def tearDown(self) -> None:
        if os.path.isdir(self._data_dir):
            shutil.rmtree(self._data_dir)

    def testProcessing_1(self):
        self.assertFalse(os.path.exists(self.tomwer_file))
        scheme_file = UtilsTest.getOrangeExecTestFile("nxtomomill_darkref.ows")
        self.scheme = self.get_scheme(scheme_file)
        _exec(scheme=self.scheme, scan=self.bliss_scan, timeout=self.TIMEOUT)
        self.assertTrue(os.path.exists(self.tomwer_file))
        with h5py.File(self.tomwer_file, "r") as h5s:
            self.assertTrue("entry0000" in h5s)
            self.assertTrue("entry0000/tomwer_process_0/results/flats" in h5s)
            self.assertTrue(
                h5s["entry0000/tomwer_process_0/results/flats/100"].shape, (2560, 2160)
            )
            self.assertFalse("entry0000/tomwer_process_0/results/darks" in h5s)

    def testProcessing_2(self):
        self.assertFalse(os.path.exists(self.tomwer_file))
        scheme_file = UtilsTest.getOrangeExecTestFile("nxtomomill_darkref_2.ows")
        self.scheme = self.get_scheme(scheme_file)
        _exec(scheme=self.scheme, scan=self.bliss_scan, timeout=self.TIMEOUT)
        self.assertTrue(os.path.exists(self.tomwer_file))
        with h5py.File(self.tomwer_file, "r") as h5s:
            self.assertTrue("entry0000" in h5s)
            self.assertTrue("entry0000/tomwer_process_0/results/flats" in h5s)
            self.assertTrue(
                h5s["entry0000/tomwer_process_0/results/flats/100"].shape, (2560, 2160)
            )
            self.assertTrue("entry0000/tomwer_process_0/results/darks" in h5s)
            self.assertTrue(
                h5s["entry0000/tomwer_process_0/results/darks/0"].shape, (2560, 2160)
            )


class TestDarkRefAxisNabu(unittest.TestCase):
    """
    Test the execution of the following workflow:
    dark and flat field -> center of rotation -> nabu reconstruction
    uses:

    - 'darkref_axis_nabu.ows'
        - dark: average, ref: median
        - cor: manual at 70
        - nabu: reconstruct slice 20
    - 'darkref_axis_nabu_2.ows'
        - dark: average, ref: median
        - cor: use centered (value should be ~79)
        - nabu: reconstruct middle slice (1080)
    """

    TIMEOUT = 120

    def setUp(self) -> None:
        unittest.TestCase.setUp(self)
        self._data_dir = tempfile.mkdtemp()
        self.master_file = os.path.join(self._data_dir, "frm_edftomomill_twoentries.nx")
        self.tomwer_file = os.path.join(self._data_dir, "tomwer_processes.h5")

        shutil.copyfile(
            UtilsTest.getH5Dataset(folderID="frm_edftomomill_twoentries.nx"),
            self.master_file,
        )
        self.scan = HDF5TomoScan(scan=self.master_file, entry="entry0000")
        self._nabu_cfg_folder = os.path.join(
            self.scan.path, nabu_settings.NABU_CFG_FILE_FOLDER
        )

    @staticmethod
    def get_scheme(file_path):
        return OwsParser.scheme_load(file_path, load_handlers=True)

    def tearDown(self) -> None:
        if os.path.isdir(self._data_dir):
            shutil.rmtree(self._data_dir)

    def testProcessing_1(self):
        self.assertFalse(os.path.exists(self.tomwer_file))
        self.assertEqual(len(glob(self._nabu_cfg_folder + os.sep + "*.cfg")), 0)
        scheme_file = UtilsTest.getOrangeExecTestFile("darkref_axis_nabu.ows")
        self.scheme = self.get_scheme(scheme_file)
        # monkey patch parameters to use nabu in dry tun mode
        self.scheme.nodes[2].properties["_rpSetting"]["dry_run"] = True
        _exec(scheme=self.scheme, scan=self.scan, timeout=self.TIMEOUT)
        self.assertTrue(os.path.exists(self.tomwer_file))
        with h5py.File(self.tomwer_file, "r") as h5s:
            # check darks
            self.assertTrue("entry0000" in h5s)
            self.assertTrue("entry0000/tomwer_process_0/results/flats" in h5s)
            self.assertTrue(
                h5s["entry0000/tomwer_process_0/results/flats/1"].shape, (2560, 2160)
            )
            self.assertTrue("entry0000/tomwer_process_0/results/darks" in h5s)
            self.assertTrue(
                h5s["entry0000/tomwer_process_0/results/darks/0"].shape, (2560, 2160)
            )
            # check axis
            self.assertEqual(
                h5s["entry0000/tomwer_process_1/program"][()], "tomwer_axis"
            )
            self.assertAlmostEqual(
                h5s["entry0000/tomwer_process_1/results/center_of_rotation"][()],
                70.0,
                places=4,
            )

            # check nabu
            self.assertEqual(
                h5s["entry0000/tomwer_process_2/program"][()], "nabu-slices"
            )
            axp = h5s[
                "entry0000/tomwer_process_2/configuration/reconstruction/rotation_axis_position"
            ][()]
            self.assertAlmostEqual(float(axp), (10.0 + 70.0), places=4)
            # +10: moved to nabu system (half width
        # check nabu files
        self.assertEqual(len(glob(self._nabu_cfg_folder + os.sep + "*.cfg")), 2)

    def testProcessing_2(self):
        self.assertFalse(os.path.exists(self.tomwer_file))
        self.assertEqual(len(glob(self._nabu_cfg_folder + os.sep + "*.cfg")), 0)

        scheme_file = UtilsTest.getOrangeExecTestFile("darkref_axis_nabu_2.ows")
        self.scheme = self.get_scheme(scheme_file)
        # monkey patch parameters to use nabu in dry tun mode
        self.scheme.nodes[2].properties["_rpSetting"]["dry_run"] = True

        _exec(scheme=self.scheme, scan=self.scan, timeout=self.TIMEOUT)
        self.assertTrue(os.path.exists(self.tomwer_file))
        with h5py.File(self.tomwer_file, "r") as h5s:
            # check darks
            self.assertTrue("entry0000" in h5s)
            self.assertTrue("entry0000/tomwer_process_0/results/flats" in h5s)
            self.assertTrue(
                h5s["entry0000/tomwer_process_0/results/flats/1"].shape, (2560, 2160)
            )
            self.assertTrue("entry0000/tomwer_process_0/results/darks" in h5s)
            self.assertTrue(
                h5s["entry0000/tomwer_process_0/results/darks/0"].shape, (2560, 2160)
            )
            # check axis
            self.assertEqual(
                h5s["entry0000/tomwer_process_1/program"][()], "tomwer_axis"
            )
            axis_cor = h5s["entry0000/tomwer_process_1/results/center_of_rotation"][()]

            # check nabu
            self.assertEqual(
                h5s["entry0000/tomwer_process_2/program"][()], "nabu-slices"
            )
            axp = h5s[
                "entry0000/tomwer_process_2/configuration/reconstruction/rotation_axis_position"
            ][()]
            self.assertAlmostEqual(float(axp), (10.0 + axis_cor), places=4)
            # +10: moved to nabu system (half width
        # check nabu files
        self.assertEqual(len(glob(self._nabu_cfg_folder + os.sep + "*.cfg")), 2)


class TestPythonScript(unittest.TestCase):
    """
    Test the execution of the following workflow:
    Data list -> Python script test -> Python script test 2

    Data list will be ignored
    The first scripts create a 'test.txt' file at the root level
    The second scripts create a 'test2.txt' file at the root level

    Make sure the code is correctly executed and that it.
    """

    TIMEOUT = 120

    def setUp(self) -> None:
        unittest.TestCase.setUp(self)
        self._data_dir = tempfile.mkdtemp()
        self.master_file = os.path.join(self._data_dir, "frm_edftomomill_twoentries.nx")
        self.tomwer_file = os.path.join(self._data_dir, "tomwer_processes.h5")

        shutil.copyfile(
            UtilsTest.getH5Dataset(folderID="frm_edftomomill_twoentries.nx"),
            self.master_file,
        )
        self.scan = HDF5TomoScan(scan=self.master_file, entry="entry0000")

    @staticmethod
    def get_scheme(file_path):
        return OwsParser.scheme_load(file_path, load_handlers=True)

    def tearDown(self) -> None:
        if os.path.isdir(self._data_dir):
            shutil.rmtree(self._data_dir)

    def testProcessing(self):
        self.assertEqual(len(glob(self.scan.path + os.sep + "test*.txt")), 0)
        scheme_file = UtilsTest.getOrangeExecTestFile("test_tomwer_python_scripts.ows")
        self.scheme = self.get_scheme(scheme_file)
        _exec(scheme=self.scheme, scan=self.scan, timeout=self.TIMEOUT)
        self.assertEqual(len(glob(self.scan.path + os.sep + "test*.txt")), 2)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestNxTomoDarkRef, TestDarkRefAxisNabu, TestPythonScript):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
""" Not a unit test but send a simple test message error to gray log"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/01/2017"

from tomwer.web.client import OWClient
import logging

logger = logging.getLogger(__name__)


class TestGrayLog(OWClient):
    def __init__(self):
        OWClient.__init__(self)

    def sendErrorMessage(self):
        logger.error("test error message")

    def sendwarningMessage(self):
        logger.warning("test warning message")

    def sendInfoMessage(self):
        logger.info("test info message")

    def sendProcessEndedMessage(self):
        logger.processEnded("test processEnded message")


if __name__ == "__main__":
    c = TestGrayLog()
    c.sendErrorMessage()
    c.sendwarningMessage()
    c.sendInfoMessage()
    c.sendProcessEndedMessage()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "03/05/2019"


from orangecontrib.tomwer.widgets.reconstruction.AxisOW import AxisOW
from orangecontrib.tomwer.test.utils import OrangeWorflowTest
from tomwer.test.utils import skip_gui_test, skip_orange_workflows_test
from tomwer.test.utils import UtilsTest
from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.core.settings import mock_lsbram
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core import utils
from tomwer.core.utils.scanutils import MockEDF, MockHDF5
from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt
import numpy
import shutil
import tempfile
import unittest
import time
import os


@unittest.skipIf(skip_gui_test(), reason="skip gui test")
class TestOWAxis(TestCaseQt):
    """Test that the axis widget work correctly"""

    def setUp(self):
        self._window = AxisOW()
        self.recons_params = self._window.recons_params
        self.scan_path = UtilsTest.getEDFDataset("D2_H2_T2_h_")
        self._window.getAxis().mode = AxisMode.centered
        self._window.show()
        self.qWaitForWindowExposed(self._window)

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()

    def testAxisLock(self):
        """Test behavior when locking the axis position. Could not be included
        in the tomwer/gui because the lock action is only available for the OW
        """
        assert self._window.getAxis().mode in (AxisMode.centered, AxisMode.global_)
        radio_axis = self._window._widget._axisWidget._radioAxis
        main_widget = self._window._widget._axisWidget
        self.assertFalse(main_widget._controlWidget._lockBut.isLocked())
        self.assertTrue(radio_axis._controlWidget._mainWidget.isEnabled())
        self.mouseClick(main_widget._controlWidget._lockBut, qt.Qt.LeftButton)
        self.qapp.processEvents()
        self.assertTrue(main_widget._controlWidget._lockBut.isLocked())
        # when the lock button is activated we should automatically switch to
        # the manual mode
        self.assertTrue(self._window.getAxis().mode is AxisMode.manual)
        self.assertFalse(radio_axis._controlWidget._mainWidget.isEnabled())


@unittest.skipIf(skip_gui_test(), reason="skip gui test")
class TestWindowAxisComputation(TestCaseQt):
    @staticmethod
    def _long_computation(scan):
        time.sleep(5)
        return -1

    """Test that the axis widget work correctly"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._mainWindow = AxisOW()
        self.recons_params = self._mainWindow.recons_params
        self._window = self._mainWindow._widget
        self.scan_path = ScanFactory.create_scan_object(
            UtilsTest.getEDFDataset("D2_H2_T2_h_")
        )
        self._mainWindow.show()
        self.qWaitForWindowExposed(self._mainWindow)

    def tearDown(self):
        self._mainWindow.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._mainWindow.close()
        self._mainWindow = None
        self._window = None
        TestCaseQt.tearDown(self)

    def testFailedComputation(self):
        """Test gui if the axis position fails"""
        self.qapp.processEvents()
        self.recons_params.set_position_frm_par_file("not existing", force=True)
        self.assertEqual(self.recons_params.mode, AxisMode.read)
        radio_axis = self._window._axisWidget._radioAxis
        self.assertEqual(radio_axis.getMode(), AxisMode.read)
        self._mainWindow.process(self.scan_path)
        self.assertEqual(radio_axis.getMode(), AxisMode.read)
        self.assertEqual(self.recons_params.value_ref_tomwer, None)
        self.assertEqual(self.recons_params.mode, AxisMode.read)
        position_info_widget = self._window._axisWidget._controlWidget._positionInfo
        self.assertEqual(position_info_widget._tomwerPositionLabel.text(), "?")

    def testLongComputation(self):
        """Test behavior if some time consuming axis computation is made"""
        # monkey patch the method used
        radioAxis = self._mainWindow._widget._axisWidget._radioAxis
        self._mainWindow._processingStack._computationThread.patch_calc_method(
            AxisMode.centered, TestWindowAxisComputation._long_computation
        )
        self.recons_params.mode = AxisMode.centered
        self.qapp.processEvents()
        self.assertTrue(radioAxis.getMode() == AxisMode.centered)
        self._mainWindow.process(self.scan_path)
        self.qapp.processEvents()
        # need the data to be loaded from thread and the cor calculation thread
        # to be launch
        time.sleep(0.5)
        self.qapp.processEvents()
        self.assertFalse(radioAxis._controlWidget._mainWidget.isEnabled())
        position_info_widget = self._window._axisWidget._controlWidget._positionInfo
        # self.assertEqual(position_info_widget._positionLabel.text(),
        #                  '...')

        # then wait for the computation end
        self._mainWindow._processingStack.wait_computation_finished()
        self.qapp.processEvents()
        time.sleep(0.5)
        self.qapp.processEvents()

        # check result is correctly computed
        radio_axis = self._window._axisWidget._radioAxis
        self.assertTrue(radio_axis._controlWidget._mainWidget.isEnabled())
        res = position_info_widget._tomwerPositionLabel.text()
        self.assertTrue(float(res) == -1.0)
        self.assertTrue(self.recons_params.value_ref_tomwer, -1.0)

    def testComputationSucceed(self):
        """Test gui if the axis position is correctly computed"""
        self.recons_params.mode = AxisMode.manual
        self.recons_params.set_value_ref_tomwer(2.345)
        self._mainWindow.process(self.scan_path)
        self.qapp.processEvents()
        radio_axis = self._window._axisWidget._radioAxis
        self.assertEqual(radio_axis.getMode(), AxisMode.manual)
        self.assertEqual(self.recons_params.value_ref_tomwer, 2.345)
        position_info_widget = self._window._axisWidget._controlWidget._positionInfo
        self.assertEqual(position_info_widget._tomwerPositionLabel.text(), "2.345")


global _computation_res
_computation_res = 0


@unittest.skipIf(skip_gui_test(), reason="skip gui test")
class TestAxisStack(TestCaseQt):
    """Test axis computation of a stack of scan"""

    @staticmethod
    def _test_computation(scan):
        global _computation_res
        _computation_res += 1
        return _computation_res

    def setUp(self):
        # not working due to OW
        TestCaseQt.setUp(self)
        self._scan1 = MockEDF.mockScan(scanID=tempfile.mkdtemp())
        self._scan2 = MockEDF.mockScan(scanID=tempfile.mkdtemp())
        self._scan3 = MockEDF.mockScan(scanID=tempfile.mkdtemp())
        self._mainWindow = AxisOW()
        self.recons_params = self._mainWindow._axis_params
        self._mainWindow._skip_exec(True)

        global _computation_res
        _computation_res = 0

        self._mainWindow.patch_calc_method(
            AxisMode.centered, TestAxisStack._test_computation
        )
        self.recons_params.mode = AxisMode.centered

        self._mainWindow.show()
        self.qWaitForWindowExposed(self._mainWindow)

    def tearDown(self):
        for scan in (self._scan1, self._scan2, self._scan3):
            shutil.rmtree(scan.path)
        utils.mockLowMemory(False)
        mock_lsbram(False)
        self._mainWindow.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._mainWindow.close()
        self._mainWindow = None
        self._recons_params = None
        self.qapp.processEvents()
        # not working due to OW
        TestCaseQt.tearDown(self)

    def testLowMemory(self):
        """Make sure the axis computation will be skip if we are in low memory"""
        utils.mockLowMemory(True)
        mock_lsbram(True)
        for scan in (self._scan1, self._scan2, self._scan3):
            self.qapp.processEvents()
            self._mainWindow.new_data_in(scan)

        self.assertEqual(self._scan1.axis_params, None)
        self.assertEqual(self._scan2.axis_params, None)
        self.assertEqual(self._scan3.axis_params, None)

    def testUnlockStack(self):
        """Check that all axis position will be computed properly if we set a
        stack of scan"""
        self._mainWindow.recons_params.set_value_ref_tomwer(1.0)
        for scan in (self._scan1, self._scan2, self._scan3):
            self._mainWindow.process(scan)

        for i in range(5):
            self.qapp.processEvents()
            time.sleep(0.2)
            self.qapp.processEvents()

        self.assertNotEqual(self._scan1.axis_params, None)
        self.assertNotEqual(self._scan2.axis_params, None)
        self.assertNotEqual(self._scan3.axis_params, None)

    def testLockStack(self):
        """Check that axis position will be simply copy if we are in a lock
        stack"""
        self.recons_params.mode = AxisMode.manual
        position_value = 0.36
        self.recons_params.set_value_ref_tomwer(position_value)

        for scan in (self._scan1, self._scan2, self._scan3):
            self._mainWindow.process(scan)

        for i in range(5):
            self.qapp.processEvents()
            time.sleep(0.2)
            self.qapp.processEvents()

        self.assertEqual(self._scan1.axis_params.value_ref_tomwer, position_value)
        self.assertEqual(self._scan2.axis_params.value_ref_tomwer, position_value)
        self.assertEqual(self._scan3.axis_params.value_ref_tomwer, position_value)


@unittest.skipIf(skip_orange_workflows_test(), reason="skip orange workflows tests")
@unittest.skipIf(skip_gui_test(), reason="skip gui test")
class AxisInWorkflow(OrangeWorflowTest):
    """Test behavior of the axis widget with the following workflow:

       * five HDF5Scan
       * DataList -> Axis -> DataList
       * Axis is set to the 'acccurate' algorithm and the processing is lock.

    We make sure once the processing is finished each has a cor value and
    end in the second DataList.
    """

    @classmethod
    def setUpClass(cls):
        OrangeWorflowTest.setUpClass()
        # create widgets
        cls.dataList1Node = cls.addWidget(
            cls, "orangecontrib.tomwer.widgets.control.DataListOW.DataListOW"
        )
        cls.axisNode = cls.addWidget(
            cls, "orangecontrib.tomwer.widgets.reconstruction.AxisOW.AxisOW"
        )
        cls.dataList2Node = cls.addWidget(
            cls, "orangecontrib.tomwer.widgets.control.DataListOW.DataListOW"
        )
        cls.processOrangeEvents(cls)

        cls.link(cls, cls.dataList1Node, "data", cls.axisNode, "data")
        cls.link(cls, cls.axisNode, "data", cls.dataList2Node, "data")
        cls.processOrangeEvents(cls)

        cls.dataList1Widget = cls.getWidgetForNode(cls, cls.dataList1Node)
        cls.axisWidget = cls.getWidgetForNode(cls, cls.axisNode)
        cls.dataList2Widget = cls.getWidgetForNode(cls, cls.dataList2Node)

        # Set we only want to simulate the reconstruction
        cls.axisWidget.setMode("centered")
        cls.axisWidget._setModeLockFrmSettings(True)

    def setUp(self):
        OrangeWorflowTest.setUp(self)
        self._source_dir = tempfile.mkdtemp()

        def create_scan(folder_name):
            _dir = os.path.join(self._source_dir, folder_name)
            return MockHDF5(
                scan_path=_dir,
                n_ini_proj=10,
                n_proj=10,
                n_alignement_proj=2,
                create_final_ref=False,
                create_ini_dark=True,
                create_ini_ref=True,
                n_refs=1,
                dim=1000,
            ).scan

        # create scans
        self.scan_1 = create_scan("scan_1")
        self.scan_2 = create_scan("scan_2")
        self.scan_3 = create_scan("scan_3")
        self.scan_4 = create_scan("scan_4")
        self.scan_5 = create_scan("scan_5")
        self.scans = [self.scan_1, self.scan_2, self.scan_3, self.scan_4, self.scan_5]

    def tearDown(self):
        shutil.rmtree(self._source_dir)
        OrangeWorflowTest.tearDown(self)

    def test(self):
        """
        Make sure if the scan receive 5 scan in 'normal way' then it will
        compute axis for each of them and move then to the DataList
        """
        for scan in self.scans:
            self.dataList1Widget.add(scan)

        # start the workflow
        self.dataList1Widget._sendList()
        for i in range(10):
            time.sleep(0.2)
            self.app.processEvents()
        self.assertEqual(self.dataList2Widget.n_scan(), 5)


#
# @unittest.skipIf(skip_gui_test(), reason="skip gui test")
# class TestHalf(TestCaseQt):
#     """Test that the interface is updated according to the scan type"""
#
#     def setUp(self):
#         # TestCaseQt.setUp(self)
#         self._source_dir = tempfile.mkdtemp()
#
#         dir_1 = os.path.join(self._source_dir, "full_field_acq")
#         self._full_field_scan = MockHDF5(
#             scan_path=dir_1,
#             n_ini_proj=10,
#             n_proj=10,
#             n_alignement_proj=2,
#             create_final_ref=False,
#             create_ini_dark=True,
#             create_ini_ref=True,
#             n_refs=1,
#             dim=100,
#             field_of_view="Full",
#         ).scan
#         dir_2 = os.path.join(self._source_dir, "half_field_acq_1")
#         self._half_field_scan_1 = MockHDF5(
#             scan_path=dir_2,
#             n_ini_proj=10,
#             n_proj=10,
#             n_alignement_proj=2,
#             create_final_ref=False,
#             create_ini_dark=True,
#             create_ini_ref=True,
#             n_refs=1,
#             dim=100,
#             field_of_view="Half",
#             estimated_cor_frm_motor=-12,
#         ).scan
#         dir_3 = os.path.join(self._source_dir, "half_field_acq_2")
#         self._half_field_scan_2 = MockHDF5(
#             scan_path=dir_3,
#             n_ini_proj=10,
#             n_proj=10,
#             n_alignement_proj=2,
#             create_final_ref=False,
#             create_ini_dark=True,
#             create_ini_ref=True,
#             n_refs=1,
#             dim=100,
#             field_of_view="Half",
#             estimated_cor_frm_motor=None,
#         ).scan
#
#         self._window = AxisOW()
#         self.recons_params = self._window.recons_params
#         self._window.getAxis().mode = AxisMode.centered
#         self._window.show()
#         self.qWaitForWindowExposed(self._window)
#
#     def tearDown(self):
#         shutil.rmtree(self._source_dir)
#         self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
#         self._window.close()
#         # TestCaseQt.tearDown(self)
#
#     def testScenario1(self):
#         """insure behavior work when self half without estimated cor, then
#         full field acquisition, then half field with estimated cor"""
#         self.assertEqual(self._window.getRadioMode(), AxisMode.centered)
#         self._window.new_data_in(self._half_field_scan_2)
#         self.assertEqual(self._window.getRadioMode(), AxisMode.near)
#         self.assertEqual(self._window.getEstimatedCor(), 0.0)
#         self._window.new_data_in(self._full_field_scan)
#         self.assertEqual(self._window.getRadioMode(), AxisMode.centered)
#         self.assertEqual(self._window.getEstimatedCor(), 0.0)
#         self._window.new_data_in(self._half_field_scan_1)
#         self.assertEqual(self._window.getRadioMode(), AxisMode.near)
#         self.assertEqual(self._window.getEstimatedCor(), -12.0)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (
        TestOWAxis,
        TestWindowAxisComputation,
        TestAxisStack,
        AxisInWorkflow,
        # TestHalf,
    ):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "30/10/2020"


import h5py
from silx.gui import qt
from functools import partial
from tomoscan.esrf.hdf5scan import ImageKey
from silx.io.url import DataUrl
from tomoscan.io import HDF5File
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from silx.gui.dialog.DataFileDialog import DataFileDialog

from typing import Union
import logging

_logger = logging.getLogger(__name__)


class _DarkOrFlatUrl(qt.QWidget):
    def __init__(self, parent, type_: ImageKey, when: str):
        assert when in ("start", "end")
        qt.QWidget.__init__(self, parent)
        self._type = type_
        self._when = when
        self.setLayout(qt.QHBoxLayout())

        self._qle = qt.QLineEdit("", self)
        self.layout().addWidget(self._qle)

        self._optionsCB = qt.QComboBox(self)
        self._optionsCB.addItem("full dataset")
        self.layout().addWidget(self._optionsCB)

        # connect signal / slot
        self._qle.editingFinished.connect(self._tryUpdateOptions)

    def setUrl(self, url):
        """
        Define the url where the dataset can be picked
        If this is a dataset from an NXEntry containing dark / flat then we
        will propose to pick only a part of the dataset (only start / end of
         dark / flat)

        :param url:
        :return:
        """
        url = self._redirectDataPath(url, logger=_logger)

        def dataset_invalid(url):
            with HDF5File(url.file_path(), mode="r") as h5s:
                if not isinstance(h5s[url.data_path()], h5py.Dataset):
                    return True
            return False

        if url.data_path() in (None, "") or dataset_invalid(url):
            msg = qt.QMessageBox(self)
            msg.setIcon(qt.QMessageBox.Warning)
            text = "You should provide a path to a dataset or to a NXtomo entry"
            msg.setText(text)
            msg.exec_()
            return

        self._qle.setText(url.path())
        self._updateOptions(url)

    def setSerie(self, serie):
        type_name = self._type.name.lower().replace("_", " ")
        item_txt = "{} serie n°{}".format(type_name, serie)
        idx = self._optionsCB.findText(item_txt)
        if idx >= 0:
            self._optionsCB.setCurrentIndex(idx)

    def _redirectDataPath(self, url, logger=None):
        try:
            with HDF5File(url.file_path(), mode="r") as h5s:
                node = h5s[url.data_path()]

                if HDF5TomoScan.entry_is_nx_tomo(node):
                    if (
                        "detector" in node["instrument"]
                        and "data" in node["instrument"]["detector"]
                    ):
                        if logger:
                            _logger.info(
                                "NXTomo entry found. Set the directly the"
                                "detector data"
                            )
                        data_path = "/".join(
                            (url.data_path(), "instrument", "detector", "data")
                        )
                        url = DataUrl(file_path=url.file_path(), data_path=data_path)
                elif HDF5TomoScan.is_nxdetector(node):
                    if "data" in node:
                        if logger:
                            _logger.info(
                                "NX_detector entry found. Set the directly"
                                " the detector data"
                            )
                        data_path = "/".join((url.data_path, "data"))
                        url = DataUrl(file_path=url.file_path(), data_path=data_path)
        except Exception as e:
            pass
        return url

    def _tryUpdateOptions(self):
        try:
            url = DataUrl(path=self._qle.text())
            self._updateOptions(url)
        except:
            pass

    def _updateOptions(self, url: DataUrl):
        self._optionsCB.clear()
        if url is None:
            return
        assert isinstance(url, DataUrl)
        assert h5py.is_hdf5(url.file_path()), "only manage hdf5 file"
        self._optionsCB.addItem("full dataset")
        if url.data_path() == "":
            _logger.error("data path should be specify")
            return
        # if we are on a 'detector / data dataset' then we can try to reach
        # image_key information
        image_keys = self._getImageKey(url)
        if image_keys is not None:
            n_serie = self._getNSeries(
                image_key_values=image_keys, image_key_type=self._type
            )
            if n_serie is not None:
                type_name = self._type.name.lower().replace("_", " ")
                series_indexes = list(range(n_serie))
                for i_serie in series_indexes:
                    self._optionsCB.addItem("{} serie n°{}".format(type_name, i_serie))
                # if this is a end url then set to the last found entry
                if self._when == "end":
                    opt_idx = self._optionsCB.findText(
                        "{} serie n°{}".format(type_name, series_indexes[-1])
                    )
                else:
                    # else set it to the first entry
                    opt_idx = self._optionsCB.findText(
                        "{} serie n°{}".format(type_name, series_indexes[0])
                    )
                if opt_idx >= 0:
                    self._optionsCB.setCurrentIndex(opt_idx)

    def _getImageKey(self, url):
        # if we are on a 'detector / data dataset' then we can try to reach
        # image_key information
        with HDF5File(url.file_path(), mode="r") as h5s:
            dataset = h5s[url.data_path()]
            grp_parent = dataset.parent
            if grp_parent is not None and HDF5TomoScan.is_nxdetector(grp_parent):
                if "image_key" in grp_parent:
                    return grp_parent["image_key"][()]
        return None

    def _getSlices(self, image_key_values, serie_index):
        n_serie = -1
        start = None
        is_in_a_serie = False
        for i_frame, frame in enumerate(image_key_values):
            if frame == self._type.value and not is_in_a_serie:
                n_serie += 1
                is_in_a_serie = True
                if n_serie == serie_index:
                    start = i_frame
            elif frame != self._type.value and is_in_a_serie:
                is_in_a_serie = False
                if n_serie == serie_index:
                    return slice(start, i_frame)
        if start is not None:
            return slice(start, len(image_key_values))
        return None

    @staticmethod
    def _getNSeries(image_key_values, image_key_type):
        n_serie = 0
        is_in_a_serie = False
        for frame in image_key_values:
            if frame == image_key_type.value and not is_in_a_serie:
                is_in_a_serie = True
                n_serie += 1
            elif frame != image_key_type.value:
                is_in_a_serie = False
        return n_serie

    def getUrl(self):
        try:
            url_no_slices = DataUrl(path=self._qle.text())
            serie_index = self._optionsCB.currentText()
            if serie_index == "full dataset":
                slices = None
            else:
                serie_index = int(serie_index.split("n°")[1])
                slices = self._getSlices(
                    image_key_values=self._getImageKey(url_no_slices),
                    serie_index=serie_index,
                )
            data_path = url_no_slices.data_path()
            url = DataUrl(
                file_path=url_no_slices.file_path(),
                data_path=data_path,
                data_slice=list(range(slices.start, slices.stop)),
                scheme="silx",
            )
        except Exception as e:
            _logger.warning("Fail to create url. Reason is", str(e))
            return None
        else:
            return url


class DarkRefPatchWidget(qt.QWidget):
    """
    Widget to add dark and flat
    """

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())

        # start dark
        self._sdCB = qt.QCheckBox("start dark", self)
        self.layout().addWidget(self._sdCB, 0, 0, 1, 1)
        self._sdQLE = _DarkOrFlatUrl(self, ImageKey.DARK_FIELD, when="start")
        self.layout().addWidget(self._sdQLE, 0, 2, 1, 1)
        self._selectSD = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectSD, 0, 3, 1, 1)

        # start flat
        self._sfCB = qt.QCheckBox("start flat", self)
        self.layout().addWidget(self._sfCB, 1, 0, 1, 1)
        self._sfQLE = _DarkOrFlatUrl(self, ImageKey.FLAT_FIELD, when="start")
        self.layout().addWidget(self._sfQLE, 1, 2, 1, 1)
        self._selectSF = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectSF, 1, 3, 1, 1)

        # end dark
        self._edCB = qt.QCheckBox("end dark ", self)
        self.layout().addWidget(self._edCB, 2, 0, 1, 1)
        self._edQLE = _DarkOrFlatUrl(self, ImageKey.DARK_FIELD, when="end")
        self.layout().addWidget(self._edQLE, 2, 2, 1, 1)
        self._selectED = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectED, 2, 3, 1, 1)

        # end flat
        self._efCB = qt.QCheckBox("end flat", self)
        self.layout().addWidget(self._efCB, 3, 0, 1, 1)
        self._efQLE = _DarkOrFlatUrl(self, ImageKey.FLAT_FIELD, when="end")
        self.layout().addWidget(self._efQLE, 3, 2, 1, 1)
        self._selectEF = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectEF, 3, 3, 1, 1)

        # signal / slot connection
        # connect select button
        self._selectSD.released.connect(self._selectSDDataset)
        self._selectSF.released.connect(self._selectSFDataset)
        self._selectED.released.connect(self._selectEDDataset)
        self._selectEF.released.connect(self._selectEFDataset)

        # set up
        for widget_ in (
            self._sdQLE,
            self._selectSD,
            self._edQLE,
            self._selectED,
            self._sfQLE,
            self._selectSF,
            self._efQLE,
            self._selectEF,
        ):
            widget_.setEnabled(False)

        # connect checkbox buttons
        self._sdCB.toggled.connect(self._toggleSD)
        self._sfCB.toggled.connect(self._toggleSF)
        self._edCB.toggled.connect(self._toggleED)
        self._efCB.toggled.connect(self._toggleEF)

    def _selectSDDataset(self):
        self._selectDataset("start dark", self._sdQLE)

    def _selectSFDataset(self):
        self._selectDataset("start flat", self._sfQLE)

    def _selectEDDataset(self):
        self._selectDataset("end dark", self._edQLE)

    def _selectEFDataset(self):
        self._selectDataset("end flat", self._efQLE)

    def _updateEnable(self, toggled, widgets):
        for w in widgets:
            w.setEnabled(toggled)

    def _toggleSD(self, toggled):
        self._updateEnable(toggled=toggled, widgets=(self._sdQLE, self._selectSD))

    def _toggleSF(self, toggled):
        self._updateEnable(toggled=toggled, widgets=(self._sfQLE, self._selectSF))

    def _toggleED(self, toggled):
        self._updateEnable(toggled=toggled, widgets=(self._edQLE, self._selectED))

    def _toggleEF(self, toggled):
        self._updateEnable(toggled=toggled, widgets=(self._efQLE, self._selectEF))

    def _selectDataset(self, what, outputQLE):
        # we can select a dataset or an entry. If this is an entry then we
        # will copy the target dark / flat...
        # ou ajouter un from: to
        dialog = DataFileDialog()
        dialog.setWindowTitle(what)
        if dialog.exec_() and dialog.selectedDataUrl() is not None:
            old = outputQLE.blockSignals(True)
            try:
                outputQLE.setUrl(dialog.selectedDataUrl())
            except Exception as e:
                _logger.error(e)
            outputQLE.blockSignals(old)

    def getStartDarkUrl(self) -> Union[None, DataUrl]:
        """

        :return: url defined by the user for patching the start darks
        :rtype:  Union[None, DataUrl]
        """
        if self._sdCB.isChecked():
            return self._sdQLE.getUrl()
        else:
            return None

    def setStartDarkUrl(self, url, serie_index=None) -> None:
        """

        :param DataUrl url:
        :param Union[int,None] serie_index: index of the serie to set
        """
        self._sdCB.setChecked(True)
        self._sdQLE.setUrl(url)
        if serie_index is not None:
            self._sdQLE.setSerie(serie_index)

    def getStartFlatUrl(self) -> Union[None, DataUrl]:
        """

        :return: url defined by the user for patching the start flats
        :rtype:  Union[None, DataUrl]
        """
        if self._sfCB.isChecked():
            return self._sfQLE.getUrl()
        else:
            return None

    def setStartFlatUrl(self, url, serie_index=None) -> None:
        """

        :param DataUrl url:
        :param Union[int,None] serie_index: index of the serie to set
        """
        self._sfCB.setChecked(True)
        self._sfQLE.setUrl(url)
        if serie_index is not None:
            self._sfQLE.setSerie(serie_index)

    def getEndDarkUrl(self) -> Union[None, DataUrl]:
        """

        :return: url defined by the user for patching the end darks
        :rtype:  Union[None, DataUrl]
        """
        if self._edCB.isChecked():
            return self._edQLE.getUrl()
        else:
            return None

    def setEndDarkUrl(self, url, serie_index=None) -> None:
        """

        :param DataUrl url:
        :param Union[int,None] serie_index: index of the serie to set
        """
        self._edCB.setChecked(True)
        self._edQLE.setUrl(url)
        if serie_index is not None:
            self._edQLE.setSerie(serie_index)

    def getEndFlatUrl(self) -> Union[None, DataUrl]:
        """

        :return: url defined by the user for patching the end flats
        :rtype:  Union[None, DataUrl]
        """
        if self._efCB.isChecked():
            return self._efQLE.getUrl()
        else:
            return None

    def setEndFlatUrl(self, url, serie_index=None) -> None:
        """

        :param DataUrl url:
        :param Union[int,None] serie_index: index of the serie to set
        """
        self._efCB.setChecked(True)
        self._efQLE.setUrl(url)
        if serie_index is not None:
            self._efQLE.setSerie(serie_index)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/12/2016"

import logging
import os
import shutil
import tempfile
import unittest

from silx.gui import qt

from orangecontrib.tomwer.widgets.control.OldDataValidatorOW import OldDataValidatorOW
from orangecontrib.tomwer.widgets.control.DataValidatorOW import DataValidatorOW
from tomwer.gui.qtapplicationmanager import QApplicationManager
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.scanutils import MockEDF
from tomwer.test.utils import UtilsTest

_qapp = QApplicationManager()

logging.disable(logging.INFO)


class BaseTestAccumulation:
    """Test that all scans will be validated if many are send and no answer
    is give until all received"""

    NB_SCAN = 10
    """Number of scan to accumulate"""

    N_RADIO = 10
    N_RECONS = 2
    DIM_MOCK_SCAN = 10

    def _setUp(self):
        self.scans = []
        for iScan in range(TestAccumulation.NB_SCAN):
            scanID = tempfile.mkdtemp()
            scan = MockEDF.mockScan(
                scanID=scanID,
                nRadio=self.N_RADIO,
                nRecons=self.N_RECONS,
                nPagRecons=0,
                dim=self.DIM_MOCK_SCAN,
            )

            self.scans.append(scan)
        self.scanValidator = None

    def _tearDown(self):
        _qapp.processEvents()
        for f in self.scans:
            if os.path.isdir(f.path):
                shutil.rmtree(f.path)
        self.scanValidator.close()

    def test(self):
        assert self.scanValidator is not None
        for scan in self.scans:
            self.scanValidator.addScan(scan)

        _qapp.processEvents()

        self.assertEqual(
            self.scanValidator.getNScanToValidate(), TestAccumulation.NB_SCAN
        )
        for scanID in self.scans:
            self.scanValidator._validateScan(scanID)
            _qapp.processEvents()
        self.assertEqual(self.scanValidator.getNScanToValidate(), 0)


class TestAccumulation(unittest.TestCase, BaseTestAccumulation):
    """Test validation for the current data validator"""

    def setUp(self):
        BaseTestAccumulation._setUp(self)

        self.scanValidator = DataValidatorOW(None)
        self.scanValidator._warnValManualShow = True
        self.scanValidator.setAttribute(qt.Qt.WA_DeleteOnClose)

    def tearDown(self):
        BaseTestAccumulation._tearDown(self)


class TestAccumulationOld(unittest.TestCase, BaseTestAccumulation):
    def setUp(self):
        BaseTestAccumulation._setUp(self)

        self.scanValidator = OldDataValidatorOW(None)
        self.scanValidator._warnValManualShow = True
        self.scanValidator.setAttribute(qt.Qt.WA_DeleteOnClose)

    def tearDown(self):
        BaseTestAccumulation._tearDown(self)


class BaseTestValidation(unittest.TestCase):
    """Test validation for the old data validator"""

    class _ValidationReceiver(qt.QObject):
        def __init__(self):
            qt.QObject.__init__(self)
            self.counter = 0

        def count(self):
            self.counter = self.counter + 1

    @classmethod
    def setUpClass(cls):
        cls.rootdir = tempfile.mkdtemp()
        cls.dataSetIDs = "test10", "test01"
        cls.scans = []
        for scanID in cls.dataSetIDs:
            dataDir = UtilsTest.getEDFDataset(scanID)
            tmpDir = os.path.join(cls.rootdir, scanID)
            cls.scans.append(tmpDir)
            shutil.copytree(dataDir, tmpDir)
        cls.validationReceiver = TestValidation._ValidationReceiver()

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.rootdir)
        del cls.validationReceiver
        if os.path.isdir(cls.rootdir):
            shutil.rmtree(cls.rootdir)

    def tearDown(self):
        self.scanValidator.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.scanValidator.close()
        self.scanValidator = None
        unittest.TestCase.tearDown(self)

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.scanValidator = None


class TestValidation(BaseTestValidation):
    """Test validation for the latest data validation widget"""

    def setUp(self):
        BaseTestValidation.setUp(self)
        self.scanValidator = DataValidatorOW()
        self.scanValidator._warnValManualShow = True
        self.scanValidator._widget.sigScanReady.connect(self.validationReceiver.count)
        self.validationReceiver.counter = 0
        self.scanValidator.setAttribute(qt.Qt.WA_DeleteOnClose)

    def testAutomaticValidation(self):
        for scan in self.scans:
            self.scanValidator.addScan(EDFTomoScan(scan))

        self.assertTrue(self.validationReceiver.counter == 0)
        self.scanValidator.setAutomaticValidation(True)
        while _qapp.hasPendingEvents():
            _qapp.processEvents()
        self.assertTrue(self.validationReceiver.counter == 2)
        self.assertTrue(len(self.scanValidator._widget._scansToValidate) == 0)


class TestValidationOld(BaseTestValidation):
    """Test validation for the latest data validation widget"""

    class _DataValidatorWidgetPatched(OldDataValidatorOW):
        """A DataValidatorOW without the Orange send signal process"""

        scanReady = qt.Signal(TomwerScanBase)
        _warnValManualShow = True

        def _validated(self, scan):
            """Callback when the validate button is pushed"""
            assert isinstance(scan, TomwerScanBase)
            if scan is not None:
                self.Outputs.data_out.send(scan)
                self.scanReady.emit(scan)
                del self._scansToValidate[scan.path]

    def testAutomaticValidation(self):
        for scan in self.scans:
            self.scanValidator.addScan(EDFTomoScan(scan))

        self.assertTrue(self.validationReceiver.counter == 0)
        self.scanValidator.getValidationWidget("scan to treat").setChecked(False)
        while _qapp.hasPendingEvents():
            _qapp.processEvents()
        self.assertTrue(self.validationReceiver.counter == 2)
        self.assertTrue(len(self.scanValidator._scansToValidate) == 0)

    def setUp(self):
        BaseTestValidation.setUp(self)
        self.scanValidator = TestValidationOld._DataValidatorWidgetPatched()
        self.scanValidator._warnValManualShow = True
        self.scanValidator.scanReady.connect(self.validationReceiver.count)
        self.validationReceiver.counter = 0
        self.scanValidator.setAttribute(qt.Qt.WA_DeleteOnClose)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (
        TestAccumulation,
        TestAccumulationOld,
        TestValidation,
        TestValidationOld,
    ):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains gui relative to axis calculation using sinogram
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/10/2019"


from silx.gui import qt
from tomwer.core.process.reconstruction.axis.params import AxisCalculationInput
from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.gui.utils.buttons import PadlockButton
from tomwer.synctools.axis import QAxisRP
from tomwer.core.scan.scanbase import TomwerScanBase
from silx.gui.plot import Plot2D
import logging

_logger = logging.getLogger(__file__)


class SinogramAxisWindow(qt.QMainWindow):
    """
    Main window for axis calculation from sinogram
    """

    sigReadyForComputation = qt.Signal()
    """signal emitted when a computation is requested"""

    sigApply = qt.Signal()
    """signal emitted when the axis reconstruction parameters are validated"""

    sigAxisEditionLocked = qt.Signal(bool)
    """Signal emitted when the status of the reconstruction parameters edition
    change"""

    sigSinoLoadStarted = qt.Signal()
    """Signal emitted when some computation is started. For this widget
    some computation can be time consuming when creating the sinogram"""
    sigSinoLoadEnded = qt.Signal()
    """Signal emitted when a computation is ended"""

    sigLockModeChanged = qt.Signal(bool)
    """signal emitted when the lock on the mode change"""

    _MARKER_NAME = "cor"
    _MARKER_COLOR = "#ebc634"

    def __init__(self, axis, parent=None):
        qt.QMainWindow.__init__(self, parent)
        self._scan = None
        if isinstance(axis, QAxisRP):
            self._axis_params = axis
        else:
            raise TypeError("axis should be an instance of QAxisRP")

        self._plot = Plot2D(parent=self)
        self._plot.getDefaultColormap().setVRange(None, None)
        self._plot.setAxesDisplayed(False)
        self._dockOpt = qt.QDockWidget(parent=self)
        self._options = _SinogramOpts(parent=self, axis_params=axis)
        self._dockOpt.setWidget(self._options)

        self.setCentralWidget(self._plot)
        self._dockOpt.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._dockOpt)

        # prepare processing
        self._loadingThread = _LoadSinoThread()

        # expose API
        self.isModeLock = self._options.isModeLock
        self._lockModeChanged = self._options._lockModeChanged

        # connect signal / slot
        self._options.sigComputationRequested.connect(self._computationRequested)
        self._options.sigAxisEditionLocked.connect(self._editionLocked)
        self._options.sigApply.connect(self._applyRequested)
        self._options.sigLockModeChanged.connect(self._expandSigLockModeChanged)
        # self._options.sigPositionChanged.connect(self._updateCORMarker)
        self._loadingThread.finished.connect(self._sinogram_loaded)

    def _expandSigLockModeChanged(self, lock):
        self.sigLockModeChanged.emit(lock)

    def setReconsParams(self, axis):
        self._axis_params = axis
        # TODO: update according to axis paramters

    def setScan(self, scan):
        self._scan = scan
        self._options.setScan(self._scan)

    def getPlot(self):
        return self._plot

    def getSettingsWidget(self):
        return self._options

    def _updatePlot(self, sinogram):
        self._plot.addImage(data=sinogram)
        self._plot.replot()

    def _computationRequested(self):
        # first load the sinogram to display it
        if self._loadingThread.isRunning():
            _logger.warning(
                "a center of rotation is already into processing,"
                "please wait until it ends"
            )
            return

        if self._scan is None:
            return

        # update scan
        self._scan.axis_params.use_sinogram = True
        self._axis_params.sinogram_line = self._options.getRadioLine()
        self._axis_params.sinogram_subsampling = self._options.getSubsampling()
        self._axis_params.calculation_input_type = (
            self._options.getCalulationInputType()
        )
        self._scan.axis_params.sinogram_line = self._axis_params.sinogram_line
        self._scan.axis_params.sinogram_subsampling = (
            self._axis_params.sinogram_subsampling
        )
        self._scan.axis_params.calculation_input_type = (
            self._axis_params.calculation_input_type
        )
        self._options.setEnabled(False)
        self.sigSinoLoadStarted.emit()
        self._loadingThread.init(
            scan=self._scan,
            line=self._options.getRadioLine(),
            subsampling=int(self._options.getSubsampling()),
        )
        self._loadingThread.start()

    def _editionLocked(self, locked):
        self.sigAxisEditionLocked.emit(locked)

    def _applyRequested(self):
        self.sigApply.emit()

    def _sinogram_loaded(self):
        """callback when the sinogram is loaded"""
        # note: cache avoid reading data twice here.
        sinogram = self._scan.get_sinogram(
            line=self._scan.axis_params.sinogram_line,
            subsampling=self._scan.axis_params.sinogram_subsampling,
        )
        self._updatePlot(sinogram=sinogram)
        self.sigSinoLoadEnded.emit()
        self.sigReadyForComputation.emit()
        self._options.setEnabled(True)

    def setLocked(self, locked):
        if self._axis_params.mode not in (AxisMode.manual, AxisMode.read):
            self._axis_params.mode = AxisMode.manual
        self._options.setLocked(locked)

    def _setModeLockFrmSettings(self, lock):
        self._options.setLocked(lock)

    def _validated(self):
        """callback when the validate button is activated"""
        self.sigApply.emit()

    def _updateCORMarker(self, value):
        if value is None:
            try:
                self._plot.removeMarker(self._MARKER_NAME)
            except:
                pass
        else:
            # change reference:
            img = self._plot.getActiveImage(just_legend=False)
            if img:
                value = value + img.getData().shape[1] // 2
            self._plot.addXMarker(
                value, legend=self._MARKER_NAME, color=self._MARKER_COLOR, text="center"
            )


class _LoadSinoThread(qt.QThread):
    def init(self, scan, line, subsampling):
        self._scan = scan
        self._line = line
        self._subsampling = subsampling

    def run(self):
        self._scan.axis_params.use_sinogram = True
        self._scan.axis_params.sinogram_line = self._line
        self._scan.axis_params.sinogram_subsampling = self._subsampling

        try:
            self._scan.get_sinogram(line=self._line, subsampling=self._subsampling)
        except ValueError as e:
            _logger.error(e)


class _SinogramOpts(qt.QWidget):
    """
    Add axis calculation options for sinogram
    """

    sigComputationRequested = qt.Signal()
    """signal emitted when a computation is requested"""

    sigApply = qt.Signal()
    """signal emitted when the axis reconstruction parameters are validated"""

    sigAxisEditionLocked = qt.Signal(bool)
    """Signal emitted when the status of the reconstruction parameters edition
    change"""

    sigPositionChanged = qt.Signal(object)
    """Signal emitted when the cor value changed"""

    sigLockModeChanged = qt.Signal(bool)
    """Signal emitted when the lock mode change"""

    def __init__(self, parent, axis_params):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._axis = axis_params
        self._scan = None

        self._dataModeWidget = qt.QWidget(parent=self)
        self._dataModeWidget.setLayout(qt.QHBoxLayout())
        self._dataModeWidget.layout().addWidget(
            qt.QLabel("Data mode", parent=self._dataModeWidget)
        )
        self._qcbDataMode = qt.QComboBox(parent=self)
        for input_type in AxisCalculationInput:
            # paganin is not managed for sinogram
            if input_type == AxisCalculationInput.transmission_pag:
                continue
            else:
                self._qcbDataMode.addItem(input_type.name(), input_type)
        self._dataModeWidget.layout().addWidget(self._qcbDataMode)
        self._lockMethodPB = PadlockButton(parent=self._dataModeWidget)
        self._lockMethodPB.setToolTip(
            "Lock the method to compute the cor. \n"
            "This will automatically call the "
            "defined algorithm each time a scan is received."
        )
        self._dataModeWidget.layout().addWidget(self._lockMethodPB)

        self.layout().addWidget(self._dataModeWidget)

        # add line
        self._lineSelWidget = qt.QWidget(parent=self)
        self._lineSelWidget.setLayout(qt.QHBoxLayout())
        self._lineSelWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._lineSB = qt.QSpinBox(parent=self)
        self._lineSelWidget.layout().addWidget(qt.QLabel("radio line", self))
        self._lineSelWidget.layout().addWidget(self._lineSB)
        self.layout().addWidget(self._lineSelWidget)
        # TODO: add an option to show the radios from here ???

        # add subsampling option
        self._subsamplingWidget = qt.QWidget(parent=self)
        self._subsamplingWidget.setLayout(qt.QHBoxLayout())
        self._subsamplingWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._subsamplingSB = qt.QSpinBox(parent=self)
        self._subsamplingSB.setMinimum(1)
        self._subsamplingSB.setValue(4)
        self._subsamplingSB.setMaximum(16)
        self._subsamplingWidget.layout().addWidget(qt.QLabel("subsampling", self))
        self._subsamplingWidget.setToolTip(
            "if you like you can only take a "
            "subsample of the sinogram to speed up process"
        )
        self._subsamplingWidget.layout().addWidget(self._subsamplingSB)
        self.layout().addWidget(self._subsamplingWidget)

        # add spacer
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        # set up
        if self._axis.sinogram_line is None:
            self._axis.sinogram_line = 0
        else:
            self._lineSB.setValue(self._axis.sinogram_line)
        self._subsamplingSB.setValue(self._axis.sinogram_subsampling)

        # make connection
        self._qcbDataMode.currentIndexChanged.connect(self._updateInputType)

    def _needComputation(self, *arg, **kwargs):
        """callback when the radio line changed"""
        if self._scan is not None:
            self.sigComputationRequested.emit()

    def _lockModeChanged(self, lock, disable_other_mode_lock=False):
        self._lockLabel.setEnabled(not lock)
        self._lockBut.setEnabled(not lock)
        self._qcbDataMode.setEnabled(not lock)
        if disable_other_mode_lock is True:
            self.setEnabled(not lock)
        self.sigLockModeChanged.emit(lock)

    def getRadioLine(self):
        return self._lineSB.value()

    def getSubsampling(self):
        return self._subsamplingSB.value()

    def setScan(self, scan):
        """
        set the gui for this scan

        :param TomwerScanBase scan:
        """
        if scan is None:
            return
        assert isinstance(scan, TomwerScanBase)
        self.blockSignals(True)
        # update line max and value
        n_line = scan.dim_2
        if n_line is None:
            n_line = 0
        self._lineSB.setMaximum(n_line)
        if scan.axis_params.sinogram_line is not None:
            self._lineSB.setValue(scan.axis_params.sinogram_line)
        elif self._lineSB.value() == 0:
            self._lineSB.setValue(n_line // 2)

        self._scan = scan
        self.blockSignals(False)

    def _updateInputType(self):
        self._axis.calculation_input_type = self.getCalulationInputType()

    def getCalulationInputType(self, *arg, **kwargs):
        return AxisCalculationInput.from_value(self._qcbDataMode.currentText())

    def setLocked(self, locked):
        self._dataModeWidget.setEnabled(not locked)
        self._qcbDataMode.setEnabled(not locked)
        self._dataModeWidget.setEnabled(not locked)
        self._lineSelWidget.setEnabled(not locked)
        self._lineSB.setEnabled(not locked)
        self._subsamplingWidget.setEnabled(not locked)
        self.sigAxisEditionLocked.emit(locked)

    def isModeLock(self):
        return self._lockMethodPB.isChecked()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "21/07/2020"

from silx.gui import qt
from Orange.widgets import widget, gui
from Orange.widgets.widget import Input, Output
from tomwer.gui.control.datavalidator import DataValidator
from tomwer.core.scan.scanbase import TomwerScanBase
from Orange.widgets import settings
import logging

_logger = logging.getLogger(__name__)


class DataValidatorOW(widget.OWWidget):
    """a data viewer able to:

    - display slices (latest reconstructed if any)
    - display radios with or without normalization

    :param parent: the parent widget
    """

    name = "data validator"
    id = "orange.widgets.tomwer.datavalidator"
    description = """Widget displaying results of a reconstruction and asking to
        the user if he want to validate or not the reconstruction. User can also ask
        for some modification on the reconstruction parameters"""
    icon = "icons/validator.png"
    priority = 23
    category = "esrfWidgets"
    keywords = ["tomography", "file", "tomwer", "acquisition", "validation"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    _viewer_config = settings.Setting(dict())

    _warnValManualShow = False
    """
    used to know if the message to inform user about `validate manually` has
    already been displayed.
    This informative message will be show under the following conditions:

        * the scanValidator contains at least `_NB_SCAN_BF_WARN`
        * this dialog have never been showed in the current session.
    """

    _NB_SCAN_BF_WARN = 10
    """
    Limit of stored scans before displaying the informative message about
    `validate manually` checkbox
    """

    assert len(DataValidator.inputs) == 1

    class Inputs:
        data_in = Input(
            name=DataValidator.inputs[0].name,
            type=DataValidator.inputs[0].type,
            doc=DataValidator.inputs[0].doc,
        )

    assert len(DataValidator.outputs) == 2

    class Outputs:
        assert DataValidator.outputs[0].name == "change recons params"
        recons_param_changed = Output(
            name=DataValidator.outputs[0].name,
            type=DataValidator.outputs[0].type,
            doc=DataValidator.outputs[0].doc,
        )
        data_out = Output(
            name=DataValidator.outputs[1].name,
            type=DataValidator.outputs[1].type,
            doc=DataValidator.outputs[1].doc,
        )

    def __init__(self, parent=None):
        widget.OWWidget.__init__(self, parent)
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._widget = DataValidator(parent=self)
        self._layout.addWidget(self._widget)
        self._setSettings(settings=self._viewer_config)

        # connect signal / slots
        self._widget.sigChangeReconsParams.connect(self._changeReconsParamsEmiited)
        self._widget.sigScanReady.connect(self._scanReadyEmitted)
        self._widget._centralWidget.sigConfigChanged.connect(self._updateSettings)

    def close(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        widget.OWWidget.close(self)

    @Inputs.data_in
    def addScan(self, scan):
        if scan is None:
            return
        assert isinstance(scan, TomwerScanBase)
        self._widget.addScan(scan)

        # in the case the memory is full, the scan can have been already
        # validated and so not accessible
        if (
            self._warnValManualShow is False
            and len(self._widget._scansToValidate) >= self._NB_SCAN_BF_WARN
        ):
            mess = (
                "Please note that the scanValidator is actually storing %s "
                "scan(s). \n"
                "Scan need to be validated manually in order to continue "
                "the workflow processing. \n"
                "you can either validate scan manually or uncheck the "
                "`validate manually` check box." % self._NB_SCAN_BF_WARN
            )

            mess = qt.QMessageBox(self, qt.QMessageBox.Information, mess)
            mess.setModal(False)
            mess.show()
            self._warnValManualShow = True

        if self.isValidationManual():
            self.show()
            self.activateWindow()
            self.raise_()

    def isValidationManual(self):
        return self._widget.isValidationManual()

    def _changeReconsParamsEmiited(self, scan):
        self.Outputs.recons_param_changed.send(scan)

    def _scanReadyEmitted(self, scan):
        self.Outputs.data_out.send(scan)

    def _updateSettings(self):
        viewer = self._widget._centralWidget
        self._viewer_config["mode"] = viewer.getDisplayMode()
        self._viewer_config["slice_opt"] = viewer.getSliceOption()
        self._viewer_config["radio_opt"] = viewer.getRadioOption()

    def _setSettings(self, settings):
        viewer = self._widget._centralWidget
        old_state = viewer.blockSignals(True)
        if "mode" in settings:
            viewer.setDisplayMode(settings["mode"])
        if "slice_opt" in settings:
            viewer.setSliceOption(settings["slice_opt"])
        if "radio_opt" in settings:
            viewer.setRadioOption(settings["radio_opt"])
        viewer.blockSignals(old_state)

    def getNScanToValidate(self):
        return len(self._widget._scansToValidate)

    def _validateScan(self, scan):
        self._widget._validateScan(scan=scan)

    def setAutomaticValidation(self, auto):
        self._widget.setAutomaticValidation(auto)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "19/07/2018"

from silx.gui import qt
from Orange.widgets import widget, gui
from Orange.widgets.widget import Input, Output
from tomwer.web.client import OWClient
from tomwer.gui.conditions.filter import FileNameFilterWidget
from tomwer.core.scan.scanbase import TomwerScanBase
import logging

logger = logging.getLogger(__name__)


class NameFilterOW(widget.OWWidget, OWClient):
    name = "scan filter"
    id = "orange.widgets.tomwer.filterow"
    description = (
        "Simple widget which filter some data directory if the name"
        "doesn't match with the pattern defined."
    )
    icon = "icons/namefilter.svg"
    priority = 106
    category = "esrfWidgets"
    keywords = ["tomography", "selection", "tomwer", "folder", "filter"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    class Inputs:
        data_in = Input(name="data", type=TomwerScanBase)

    class Outputs:
        data_out = Output(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        """"""
        widget.OWWidget.__init__(self, parent)
        OWClient.__init__(self)

        self.widget = FileNameFilterWidget(parent=self)
        self.widget.setContentsMargins(0, 0, 0, 0)

        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self.widget)
        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        layout.addWidget(spacer)

        # expose API
        self.setPattern = self.widget.setPattern

    @Inputs.data_in
    def applyfilter(self, scan):
        if scan is None:
            return
        assert isinstance(scan, TomwerScanBase)
        if not self.widget.isFiltered(scan.path):
            self._signalScanReady(scan)

    def _signalScanReady(self, scan):
        self.Outputs.data_out.send(scan)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "19/07/2018"


import unittest
import shutil
import tempfile

from tomwer.core.utils.scanutils import MockEDF
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.process.conditions.filters import (
    RegularExpressionFilter,
    FileNameFilter,
)


class TestConditionalFilter(unittest.TestCase):
    """
    Small unit test for the core.conditions
    """

    def testPattern1(self):
        filter = RegularExpressionFilter("name10")
        self.assertTrue(filter.isFiltered("toto") is True)
        self.assertTrue(filter.isFiltered("name10") is False)
        self.assertTrue(filter.isFiltered("name100") is False)

    def testInputOutput(self):
        pass


class TestConditionIO(unittest.TestCase):
    """Test inputs and outputs types of the handler functions"""

    def setUp(self):
        self.scan_folder = tempfile.mkdtemp()

        self.scan = MockEDF.mockScan(
            scanID=self.scan_folder, nRadio=10, nRecons=1, nPagRecons=4, dim=10
        )
        self.filter_process = FileNameFilter("*")

    def tearDown(self):
        shutil.rmtree(self.scan_folder)

    def testInputOutput(self):
        for input_type in (dict, TomwerScanBase):
            for _input in FileNameFilter.inputs:
                for return_dict in (True, False):
                    with self.subTest(
                        handler=_input.handler,
                        return_dict=return_dict,
                        input_type=input_type,
                    ):
                        self.filter_process._set_return_dict(return_dict)
                        input_obj = self.scan
                        if input_type is dict:
                            input_obj = self.scan.to_dict()
                        out = getattr(self.filter_process, _input.handler)(input_obj)
                        if return_dict:
                            self.assertTrue(isinstance(out, dict))
                        else:
                            self.assertTrue(isinstance(out, TomwerScanBase))


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestConditionalFilter, TestConditionIO):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/01/2017"

import logging
import unittest
from tomwer.gui.qtapplicationmanager import QApplicationManager
from orangecontrib.tomwer.test.utils import OrangeWorflowTest
from tomwer.test.utils import skip_gui_test, skip_orange_workflows_test

logging.disable(logging.INFO)

app = QApplicationManager()


@unittest.skipIf(skip_orange_workflows_test(), reason="skip orange workflows tests")
@unittest.skipIf(skip_gui_test(), reason="skip gui test")
class TestOrangeWidgetsCreations(OrangeWorflowTest):
    """Create dummy workflow to make sure orange widget are correctly created"""

    def setUp(self):
        OrangeWorflowTest.setUp(self)

    def tearDown(self):
        OrangeWorflowTest.tearDown(self)

    def test(self):
        """Simple creation of the following workflow:
        ScanList -> ScanSelector -> nameFilter -> sampleMoved
                                               -> stackSlice
        """
        dataListNode = self.addWidget(
            "orangecontrib.tomwer.widgets.control.DataListOW.DataListOW"
        )
        dataSelectorNode = self.addWidget(
            "orangecontrib.tomwer.widgets.control.DataSelectorOW.DataSelectorOW"
        )
        scanFilterNode = self.addWidget(
            "orangecontrib.tomwer.widgets.control.FilterOW.NameFilterOW"
        )
        sampleMovedNode = self.addWidget(
            "orangecontrib.tomwer.widgets.visualization.SampleMovedOW.SampleMovedOW"
        )
        groupSliceNode = self.addWidget(
            "orangecontrib.tomwer.widgets.visualization.SliceStackOW.SlicesStackOW"
        )

        self.processOrangeEvents()

        self.link(dataListNode, "data", dataSelectorNode, "data")
        self.link(dataSelectorNode, "data", scanFilterNode, "data")
        self.link(scanFilterNode, "data", sampleMovedNode, "data")
        self.link(scanFilterNode, "data", groupSliceNode, "data")
        self.processOrangeEvents()


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestOrangeWidgetsCreations,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys
import os
from silx.gui import qt
import argparse
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.gui.edit.dkrfpatch import DarkRefPatchWidget
from tomwer.core.scan.scanfactory import ScanFactory
from nxtomomill.utils import add_dark_flat_nx_file
import signal

logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__name__)


class _DarkRefPatchDialog(qt.QDialog):
    def __init__(self, parent):
        qt.QDialog.__init__(self, parent)
        self._scan = None
        self.setLayout(qt.QVBoxLayout())
        self._mainWidget = DarkRefPatchWidget(self)
        self.layout().addWidget(self._mainWidget)

        # buttons
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self._buttons.button(qt.QDialogButtonBox.Ok).setText("Apply")
        self.layout().addWidget(self._buttons)

        # connect signal slot
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self._apply)

    def _apply(self):
        if self._scan is None:
            raise ValueError("Scan should be set before applying patch")
        url_sd = self._mainWidget.getStartDarkUrl()
        url_sf = self._mainWidget.getStartFlatUrl()
        url_ed = self._mainWidget.getEndDarkUrl()
        url_ef = self._mainWidget.getEndFlatUrl()
        add_dark_flat_nx_file(
            file_path=self._scan.master_file,
            entry=self._scan.entry,
            darks_start=url_sd,
            flats_start=url_sf,
            darks_end=url_ed,
            flats_end=url_ef,
        )

    def setScan(self, scan):
        self._scan = scan


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path",
        help="HDF5 - Nexus file path",
        default=None,
    )
    parser.add_argument("entry", help="NXTomo entry to load", default=None)
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )

    options = parser.parse_args(argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication([])
    splash = getMainSplashScreen()
    qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    if options.scan_path is not None:
        if os.path.isdir(options.scan_path):
            options.scan_path = options.scan_path.rstrip(os.path.sep)
            scan = ScanFactory.create_scan_object(scan_path=options.scan_path)
        else:
            if not os.path.isfile(options.scan_path):
                raise ValueError(
                    "scan path should be a folder containing an"
                    " EDF acquisition or an hdf5 - nexus "
                    "compliant file"
                )
            if options.entry is None:
                raise ValueError("entry in the master file should be specify")
            scan = HDF5TomoScan(scan=options.scan_path, entry=options.entry)
    else:
        scan = ScanFactory.mock_scan()
        if options.radio1 is None:
            raise ValueError("At least a file or a scan path should be given")
    # define the process_index is any tomwer_processes_existing
    scan.set_process_index_frm_tomwer_process_file()

    if options.debug:
        _logger.setLevel(logging.DEBUG)

    window = _DarkRefPatchDialog(parent=None)
    window.setScan(scan=scan)
    window.setWindowTitle("patch dark & flat")
    window.setWindowIcon(icons.getQIcon("tomwer"))

    splash.finish(window)
    window.show()
    qt.QApplication.restoreOverrideCursor()
    app.exec_()


def getinputinfo():
    return "tomwer darkref-patch [scanDir]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "09/06/2020"


import unittest
import requests
import tempfile
import shutil
import os
import h5py
from tomwer.core.process.control.datalistener.rpcserver import _BaseDataListenerThread


class JSONRPCClient:
    """Simulate call from bliss"""

    def scan_started(self, scan_number):
        return {
            "method": "scan_started",
            "params": [scan_number],
            "jsonrpc": "2.0",
            "id": 0,
        }

    def scan_ended(self, scan_number):
        return {
            "method": "scan_ended",
            "params": [scan_number],
            "jsonrpc": "2.0",
            "id": 1,
        }

    def sequence_started(
        self, saving_file, scan_title, sequence_scan_number, proposal_file, sample_file
    ):
        params = [
            saving_file,
            scan_title,
            sequence_scan_number,
            proposal_file,
            sample_file,
        ]
        return {
            "method": "sequence_started",
            "params": params,
            "jsonrpc": "2.0",
            "id": 2,
        }

    def sequence_ended(self, saving_file, sequence_scan_number, success):
        return {
            "method": "sequence_ended",
            "params": [saving_file, sequence_scan_number, success],
            "jsonrpc": "2.0",
            "id": 3,
        }

    def execute(self):
        url = "http://localhost:4000/jsonrpc"
        timeout = 1.5

        requests.post(
            url,
            json=self.sequence_started(
                saving_file="saving_file.h5",
                scan_title="my scan title",
                entry_name="entry0000",
                proposal_file="ihpropfile.h5",
            ),
            timeout=timeout,
        ).json()
        requests.post(
            url, json=self.scan_started(scan_number="0001"), timeout=timeout
        ).json()
        requests.post(
            url, json=self.scan_ended(scan_number="0001"), timeout=timeout
        ).json()
        requests.post(
            url,
            json=self.sequence_ended(
                saving_file="saving_file.h5", entry_name="entry0000", suceed=True
            ),
            timeout=timeout,
        ).json()


class TestJsonRPCServer(unittest.TestCase):
    """test the json rpc server"""

    def setUp(self):
        self.input_dir = tempfile.mkdtemp()
        sample_dir = os.path.join(self.input_dir, "sample")
        os.mkdir(sample_dir)
        self._proposal_file = os.path.join(sample_dir, "ihpayno_sample.h5")
        with h5py.File(self._proposal_file, mode="w") as h5f:
            h5f["test"] = "toto"

        os.mkdir(os.path.join(sample_dir, "sample_29042021"))
        self._sample_file = os.path.join(
            sample_dir, "sample_29042021", "sample_29042021.h5"
        )
        self._sample_file_entry = "1.1"

        with h5py.File(self._sample_file, mode="w") as h5f:
            h5f[self._sample_file_entry] = "tata"

        self.data_listener = _BaseDataListenerThread(host="localhost", port=4000)

    def tearDown(self):
        shutil.rmtree(self.input_dir)
        self.data_listener.stop()

    def testStartStop(self):
        self.data_listener.start()
        self.data_listener.stop()


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestJsonRPCServer,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "16/08/2018"

from silx.gui import qt
from tomwer.gui import icons
import numpy
import logging

_logger = logging.getLogger(__name__)


class SelectionLineEdit(qt.QWidget):
    """Line edit with several type of selection possible:

    * a single value
    * a range of value on the type min:max:step
    * a list of value: val1, val2, ...
    """

    # SINGLE_MODE = 'single'
    RANGE_MODE = "range"
    LIST_MODE = "list"

    # SELECTION_MODES = (SINGLE_MODE, RANGE_MODE, LIST_MODE)
    SELECTION_MODES = (RANGE_MODE, LIST_MODE)

    _DEFAULT_SELECTION = LIST_MODE

    def __init__(self, text=None, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self._qLineEdit = qt.QLineEdit(parent=self)
        fpm = "\\d*\\.?\\d+"  # float or int matching
        qRegExp = qt.QRegExp(
            "(" + fpm + "[;]?[,]?[ ]?){1,}" + "|" + ":".join((fpm, fpm, fpm))
        )
        self._qLineEdit.setValidator(qt.QRegExpValidator(qRegExp))
        self.layout().addWidget(self._qLineEdit)
        self._button = SelectionModeButton(parent=self)
        self.layout().addWidget(self._button)

        # Qobject signal connections
        self._qLineEdit.textChanged.connect(self._checkIfModeChanged)
        self._button.sigModeChanged.connect(self._modeChanged)

        # expose API
        self.setText = self._qLineEdit.setText
        self.editingFinished = self._qLineEdit.editingFinished
        self.textChanged = self._qLineEdit.textChanged
        self.text = self._qLineEdit.text

        if text is not None:
            self._qLineEdit.setText(str(text))

    def getMode(self):
        return self._button.mode

    @property
    def selection(self):
        if self._qLineEdit.hasAcceptableInput():
            if self._button.mode == self.RANGE_MODE:
                _from, _to, _step = self._qLineEdit.text().split(":")
                _from, _to, _step = float(_from), float(_to), float(_step)
                if _from > _to:
                    _logger.warning("to > from, invert %s and %s" % (_from, _to))
                    tmp = _to
                    _to = _from
                    _from = tmp
                num = int((_to - _from) / _step)
                return tuple(
                    numpy.linspace(start=_from, stop=_to, num=num, endpoint=True)
                )
            else:
                vals = self._qLineEdit.text().replace(" ", "")
                vals = vals.replace(";", ",").split(",")
                res = []
                [res.append(float(val)) for val in vals]
                if len(res) == 1:
                    return res[0]
                else:
                    return tuple(res)
        else:
            _logger.warning("Wrong input, unvalid selection")
            return None

    def _checkIfModeChanged(self, _str):
        self._button.blockSignals(True)
        if _str.count(":") > 0:
            self._button.mode = self.RANGE_MODE
        else:
            self._button.mode = self.LIST_MODE
        self._button.blockSignals(False)

    def _modeChanged(self, mode):
        if mode == self.RANGE_MODE:
            text = "from:to:step"
        elif mode == self.LIST_MODE:
            text = "val1; val2; ..."
        else:
            raise ValueError("unknow mode")

        self._qLineEdit.blockSignals(True)
        self._qLineEdit.setText(text)
        self._qLineEdit.selectAll()
        self._qLineEdit.blockSignals(False)


class SelectionModeButton(qt.QToolButton):
    """Base class for Selection QAction.

    :param str mode: the mode of selection of the action.
    :param str text: The name of this action to be used for menu label
    :param str tooltip: The text of the tooltip
    :param triggered: The callback to connect to the action's triggered
                      signal or None for no callback.
    """

    sigModeChanged = qt.Signal(str)

    def __init__(self, parent=None, tooltip=None, triggered=None):
        qt.QToolButton.__init__(self, parent)
        self._states = {}
        self._mode = None
        for mode in SelectionLineEdit.SELECTION_MODES:
            icon = icons.getQIcon("_".join([mode, "selection"]))
            self._states[mode] = (icon, self._getTooltip(mode))

        self._rangeAction = RangeSelAction(parent=self)
        self._listAction = ListSelAction(parent=self)
        for _action in (self._rangeAction, self._listAction):
            _action.sigModeChanged.connect(self._modeChanged)

        menu = qt.QMenu(self)
        menu.addAction(self._rangeAction)
        menu.addAction(self._listAction)
        self.setMenu(menu)
        self.setPopupMode(qt.QToolButton.InstantPopup)
        self.mode = SelectionLineEdit.LIST_MODE

    def _getTooltip(self, mode):
        # if mode == SelectionLineEdit.SINGLE_MODE:
        #     return 'Define only one value for this parameter'
        if mode == SelectionLineEdit.LIST_MODE:
            return (
                "Define a single value or a list of values for this "
                "parameter (va1; val2)"
            )
        elif mode == SelectionLineEdit.RANGE_MODE:
            return "Define a range of value for this parameter (from:to:step)"
        else:
            raise ValueError("unknow mode")

    def _modeChanged(self, mode):
        self.mode = mode

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        assert mode in SelectionLineEdit.SELECTION_MODES
        if mode != self._mode:
            self._mode = mode
            self.setIcon(icons.getQIcon("_".join([mode, "selection"])))
            self.setToolTip(self._getTooltip(mode))
            self.sigModeChanged.emit(self._mode)


class SelectionAction(qt.QAction):
    """
    Base class of the several selection mode
    """

    sigModeChanged = qt.Signal(str)

    def __init__(self, mode, parent, text):
        icon = icons.getQIcon("_".join([mode, "selection"]))
        qt.QAction.__init__(self, icon, text, parent)
        self.setIconVisibleInMenu(True)
        self._mode = mode
        self.triggered.connect(self._modeChanged)

    def _modeChanged(self, *args, **kwargs):
        self.sigModeChanged.emit(self._mode)


class RangeSelAction(SelectionAction):
    """
    Action to select a range of element on the scheme from:to:step
    """

    def __init__(self, parent=None):
        SelectionAction.__init__(
            self,
            mode=SelectionLineEdit.RANGE_MODE,
            parent=parent,
            text="range selection",
        )


class ListSelAction(SelectionAction):
    """
    Action to select a list of element on the scheme elmt1, elmt2, ...
    """

    def __init__(self, parent=None):
        SelectionAction.__init__(
            self, mode=SelectionLineEdit.LIST_MODE, parent=parent, text="list selection"
        )


if __name__ == "__main__":
    qpp = qt.QApplication([])
    w = SelectionLineEdit(None)
    w.show()
    qpp.exec_()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/10/2019"

from tomwer.test.utils import skip_gui_test, skip_orange_workflows_test
from orangecontrib.tomwer.test.utils import OrangeWorflowTest
from silx.gui.utils.testutils import SignalListener
from tomwer.core.utils.scanutils import MockEDF
from tomwer.core.scan.scanfactory import ScanFactory
import unittest
import tempfile
import shutil
import time
import os


@unittest.skipIf(skip_gui_test(), reason="skip gui test")
@unittest.skipIf(skip_orange_workflows_test(), reason="skip orange workflows tests")
class TestCopyNFolder(OrangeWorflowTest):
    """test the following workflow and behavior.
    Workflow is :
        - DataListOW
        - FTSerieWidget
        - DataTransfertOW

    A list of folder into ScanList them go through FTserieWidget and to
    FolderTransfert. Make sure all the data have been copied
    """

    TIMEOUT_TEST = 20

    @classmethod
    def setUpClass(cls):
        OrangeWorflowTest.setUpClass()
        # create widgets
        cls.scanListNode = cls.addWidget(
            cls, "orangecontrib.tomwer.widgets.control.DataListOW.DataListOW"
        )
        cls.FTSerieNode = cls.addWidget(
            cls, "orangecontrib.tomwer.widgets.reconstruction.FtseriesOW.FtseriesOW"
        )
        cls.folderTransfertNode = cls.addWidget(
            cls, "orangecontrib.tomwer.widgets.control.DataTransfertOW.DataTransfertOW"
        )

        # Let Orange process events (node creations)
        cls.processOrangeEvents(cls)

        # linking the workflow
        cls.link(cls, cls.scanListNode, "data", cls.FTSerieNode, "data")
        cls.link(cls, cls.FTSerieNode, "data", cls.folderTransfertNode, "data")

        # Let Orange process events (node creations)
        cls.processOrangeEvents(cls)

        # getting the widgets
        cls.scanListWidget = cls.getWidgetForNode(cls, cls.scanListNode)
        cls.ftserieWidget = cls.getWidgetForNode(cls, cls.FTSerieNode)
        cls.transfertWidget = cls.getWidgetForNode(cls, cls.folderTransfertNode)

        cls.processOrangeEvents(cls)

        # set mock mode for FTSerieWidget
        cls.ftserieWidget._ftserie.reconsStack.setMockMode(True)
        cls.ftserieWidget._ftserie.setForceSync(True)
        cls.transfertWidget.turn_off_print = True
        # to avoid one folder transfert to use a thread
        cls.transfertWidget.setForceSync(True)
        # add signal listener
        cls.dataTransfertListener = SignalListener()
        cls.transfertWidget.scanready.connect(cls.dataTransfertListener)

    @classmethod
    def tearDownClass(cls):
        for node in (cls.scanListNode, cls.FTSerieNode, cls.folderTransfertNode):
            cls.removeNode(cls, node)
        cls.scanListWidget = None
        cls.ftserieWidget = None
        cls.transfertWidget = None
        OrangeWorflowTest.tearDownClass()

    def setUp(self):
        OrangeWorflowTest.setUp(self)

        # reset signal listener
        self.dataTransfertListener.clear()

        # Create n folder
        self.folders = []
        for i in range(1):
            inputFolder = tempfile.mkdtemp()
            self.folders.append(inputFolder)
            # copy files directly
            MockEDF.fastMockAcquisition(inputFolder)

        self.outputdir = tempfile.mkdtemp()

    def tearDown(self):
        # remove input folders if any
        for f in self.folders:
            if os.path.isdir(f):
                shutil.rmtree(f)

        # remove output folders if any
        if os.path.isdir(self.outputdir):
            shutil.rmtree(self.outputdir)

        OrangeWorflowTest.tearDown(self)

    def testCopy(self):
        # add outputdir to transfertFolderWidget
        self.assertTrue(os.path.isdir(self.outputdir))
        self.transfertWidget.setDestDir(self.outputdir)

        # add all fodler into the scanList
        for f in self.folders:
            assert os.path.isdir(f)
            self.scanListWidget.add(ScanFactory.create_scan_object(scan_path=f))
            self.app.processEvents()
            self.processOrangeEventsStack()

        # make sure nothing has been copied
        self.assertTrue(self.outpudirIsEmpty())

        # then start workflow run by asking the scanListWidget to notice action
        # to the next widget
        self.scanListWidget.start()

        timeout = self.TIMEOUT_TEST
        while timeout > 0 and self.dataTransfertListener.callCount() == 0:
            timeout = timeout - 0.2
            self.app.processEvents()
            self.processOrangeEventsStack()
            time.sleep(0.2)

        self.assertTrue(self.dataHasBeenCopied())
        self.app.processEvents()

    def outpudirIsEmpty(self):
        return len(os.listdir(self.outputdir)) == 0

    def dataHasBeenCopied(self):
        return len(os.listdir(self.outputdir)) == 1


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestCopyNFolder,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "25/05/2018"

from Orange.widgets import widget, gui
from Orange.widgets.widget import Output, Input
from tomwer.web.client import OWClient
from tomwer.gui.scanselectorwidget import ScanSelectorWidget
from tomwer.core.scan.scanbase import TomwerScanBase
import logging

logger = logging.getLogger(__name__)


class DataSelectorOW(widget.OWWidget, OWClient):
    name = "data selector"
    id = "orange.widgets.tomwer.scanselector"
    description = (
        "List all received scan. Then user can select a specific"
        "scan to be passed to the next widget."
    )
    icon = "icons/scanselector.svg"
    priority = 42
    category = "esrfWidgets"
    keywords = ["tomography", "selection", "tomwer", "folder"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    class Inputs:
        data_in = Input(name="data", type=TomwerScanBase)

    class Outputs:
        data_out = Output(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        """"""
        widget.OWWidget.__init__(self, parent)
        OWClient.__init__(self)

        self.widget = ScanSelectorWidget(parent=self)
        self.widget.sigSelectionChanged.connect(self.changeSelection)
        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self.widget)
        # expose API
        self.setActiveScan = self.widget.setActiveScan
        self.selectAll = self.widget.selectAll
        self.n_scan = self.widget.n_scan
        self.add = self.widget.add

    @Inputs.data_in
    def addScan(self, scan):
        if scan is not None:
            self.widget.add(scan)

    def changeSelection(self, list_scan):
        if list_scan:
            for scan_id in list_scan:
                if scan_id in self.widget.dataList._scanIDs:
                    scan = self.widget.dataList._scanIDs[scan_id]
                    assert isinstance(scan, TomwerScanBase)
                    self.Outputs.data_out.send(scan)
                else:
                    logger.error("%s not found in scan ids" % scan_id)

    def send(self):
        """send output signals for each selected items"""
        sItem = self.widget.dataList.selectedItems()
        if sItem and len(sItem) >= 1:
            selection = [_item.text() for _item in sItem]
            self.changeSelection(list_scan=selection)


if __name__ == "__main__":
    from silx.gui import qt

    qapp = qt.QApplication([])
    widget = DataSelectorOW(parent=None)
    widget.show()
    qapp.exec_()
